import sys
import os
import unittest
from unittest import mock

from scrapy.item import Item, Field
from scrapy.utils.misc import arg_to_iter, create_instance, load_object, rel_has_nofollow, set_environ, walk_modules


__doctests__ = ['scrapy.utils.misc']


class UtilsMiscTestCase(unittest.TestCase):

    def test_load_object_class(self):
        obj = load_object(Field)
        self.assertIs(obj, Field)
        obj = load_object('scrapy.item.Field')
        self.assertIs(obj, Field)

    def test_load_object_function(self):
        obj = load_object(load_object)
        self.assertIs(obj, load_object)
        obj = load_object('scrapy.utils.misc.load_object')
        self.assertIs(obj, load_object)

    def test_load_object_exceptions(self):
        self.assertRaises(ImportError, load_object, 'nomodule999.mod.function')
        self.assertRaises(NameError, load_object, 'scrapy.utils.misc.load_object999')
        self.assertRaises(TypeError, load_object, {})

    def test_walk_modules(self):
        mods = walk_modules('tests.test_utils_misc.test_walk_modules')
        expected = [
            'tests.test_utils_misc.test_walk_modules',
            'tests.test_utils_misc.test_walk_modules.mod',
            'tests.test_utils_misc.test_walk_modules.mod.mod0',
            'tests.test_utils_misc.test_walk_modules.mod1',
        ]
        self.assertEqual({m.__name__ for m in mods}, set(expected))

        mods = walk_modules('tests.test_utils_misc.test_walk_modules.mod')
        expected = [
            'tests.test_utils_misc.test_walk_modules.mod',
            'tests.test_utils_misc.test_walk_modules.mod.mod0',
        ]
        self.assertEqual({m.__name__ for m in mods}, set(expected))

        mods = walk_modules('tests.test_utils_misc.test_walk_modules.mod1')
        expected = [
            'tests.test_utils_misc.test_walk_modules.mod1',
        ]
        self.assertEqual({m.__name__ for m in mods}, set(expected))

        self.assertRaises(ImportError, walk_modules, 'nomodule999')

    def test_walk_modules_egg(self):
        egg = os.path.join(os.path.dirname(__file__), 'test.egg')
        sys.path.append(egg)
        try:
            mods = walk_modules('testegg')
            expected = [
                'testegg.spiders',
                'testegg.spiders.a',
                'testegg.spiders.b',
                'testegg'
            ]
            self.assertEqual({m.__name__ for m in mods}, set(expected))
        finally:
            sys.path.remove(egg)

    def test_arg_to_iter(self):

        class TestItem(Item):
            name = Field()

        assert hasattr(arg_to_iter(None), '__iter__')
        assert hasattr(arg_to_iter(100), '__iter__')
        assert hasattr(arg_to_iter('lala'), '__iter__')
        assert hasattr(arg_to_iter([1, 2, 3]), '__iter__')
        assert hasattr(arg_to_iter(c for c in 'abcd'), '__iter__')

        self.assertEqual(list(arg_to_iter(None)), [])
        self.assertEqual(list(arg_to_iter('lala')), ['lala'])
        self.assertEqual(list(arg_to_iter(100)), [100])
        self.assertEqual(list(arg_to_iter(c for c in 'abc')), ['a', 'b', 'c'])
        self.assertEqual(list(arg_to_iter([1, 2, 3])), [1, 2, 3])
        self.assertEqual(list(arg_to_iter({'a': 1})), [{'a': 1}])
        self.assertEqual(list(arg_to_iter(TestItem(name="john"))), [TestItem(name="john")])

    def test_create_instance(self):
        settings = mock.MagicMock()
        crawler = mock.MagicMock(spec_set=['settings'])
        args = (True, 100.)
        kwargs = {'key': 'val'}

        def _test_with_settings(mock, settings):
            create_instance(mock, settings, None, *args, **kwargs)
            if hasattr(mock, 'from_crawler'):
                self.assertEqual(mock.from_crawler.call_count, 0)
            if hasattr(mock, 'from_settings'):
                mock.from_settings.assert_called_once_with(settings, *args,
                                                           **kwargs)
                self.assertEqual(mock.call_count, 0)
            else:
                mock.assert_called_once_with(*args, **kwargs)

        def _test_with_crawler(mock, settings, crawler):
            create_instance(mock, settings, crawler, *args, **kwargs)
            if hasattr(mock, 'from_crawler'):
                mock.from_crawler.assert_called_once_with(crawler, *args,
                                                          **kwargs)
                if hasattr(mock, 'from_settings'):
                    self.assertEqual(mock.from_settings.call_count, 0)
                self.assertEqual(mock.call_count, 0)
            elif hasattr(mock, 'from_settings'):
                mock.from_settings.assert_called_once_with(settings, *args,
                                                           **kwargs)
                self.assertEqual(mock.call_count, 0)
            else:
                mock.assert_called_once_with(*args, **kwargs)

        # Check usage of correct constructor using four mocks:
        #   1. with no alternative constructors
        #   2. with from_settings() constructor
        #   3. with from_crawler() constructor
        #   4. with from_settings() and from_crawler() constructor
        spec_sets = (
            ['__qualname__'],
            ['__qualname__', 'from_settings'],
            ['__qualname__', 'from_crawler'],
            ['__qualname__', 'from_settings', 'from_crawler'],
        )
        for specs in spec_sets:
            m = mock.MagicMock(spec_set=specs)
            _test_with_settings(m, settings)
            m.reset_mock()
            _test_with_crawler(m, settings, crawler)

        # Check adoption of crawler settings
        m = mock.MagicMock(spec_set=['__qualname__', 'from_settings'])
        create_instance(m, None, crawler, *args, **kwargs)
        m.from_settings.assert_called_once_with(crawler.settings, *args,
                                                **kwargs)

        with self.assertRaises(ValueError):
            create_instance(m, None, None)

        m.from_settings.return_value = None
        with self.assertRaises(TypeError):
            create_instance(m, settings, None)

    def test_set_environ(self):
        assert os.environ.get('some_test_environ') is None
        with set_environ(some_test_environ='test_value'):
            assert os.environ.get('some_test_environ') == 'test_value'
        assert os.environ.get('some_test_environ') is None

        os.environ['some_test_environ'] = 'test'
        assert os.environ.get('some_test_environ') == 'test'
        with set_environ(some_test_environ='test_value'):
            assert os.environ.get('some_test_environ') == 'test_value'
        assert os.environ.get('some_test_environ') == 'test'

    def test_rel_has_nofollow(self):
        assert rel_has_nofollow('ugc nofollow') is True
        assert rel_has_nofollow('ugc,nofollow') is True
        assert rel_has_nofollow('ugc') is False
        assert rel_has_nofollow('nofollow') is True
        assert rel_has_nofollow('nofollowfoo') is False
        assert rel_has_nofollow('foonofollow') is False
        assert rel_has_nofollow('ugc,  ,  nofollow') is True


if __name__ == "__main__":
    unittest.main()



from testfixtures import LogCapture
from twisted.internet import defer
from twisted.trial.unittest import TestCase

from scrapy.http import Request
from scrapy.crawler import CrawlerRunner
from tests.spiders import MockServerSpider
from tests.mockserver import MockServer


class InjectArgumentsDownloaderMiddleware:
    """
    Make sure downloader middlewares are able to update the keyword arguments
    """
    def process_request(self, request, spider):
        if request.callback.__name__ == 'parse_downloader_mw':
            request.cb_kwargs['from_process_request'] = True
        return None

    def process_response(self, request, response, spider):
        if request.callback.__name__ == 'parse_downloader_mw':
            request.cb_kwargs['from_process_response'] = True
        return response


class InjectArgumentsSpiderMiddleware:
    """
    Make sure spider middlewares are able to update the keyword arguments
    """
    def process_start_requests(self, start_requests, spider):
        for request in start_requests:
            if request.callback.__name__ == 'parse_spider_mw':
                request.cb_kwargs['from_process_start_requests'] = True
            yield request

    def process_spider_input(self, response, spider):
        request = response.request
        if request.callback.__name__ == 'parse_spider_mw':
            request.cb_kwargs['from_process_spider_input'] = True
        return None

    def process_spider_output(self, response, result, spider):
        for element in result:
            if isinstance(element, Request) and element.callback.__name__ == 'parse_spider_mw_2':
                element.cb_kwargs['from_process_spider_output'] = True
            yield element


class KeywordArgumentsSpider(MockServerSpider):
    name = 'kwargs'
    custom_settings = {
        'DOWNLOADER_MIDDLEWARES': {
            InjectArgumentsDownloaderMiddleware: 750,
        },
        'SPIDER_MIDDLEWARES': {
            InjectArgumentsSpiderMiddleware: 750,
        },
    }

    checks = []

    def start_requests(self):
        data = {'key': 'value', 'number': 123, 'callback': 'some_callback'}
        yield Request(self.mockserver.url('/first'), self.parse_first, cb_kwargs=data)
        yield Request(self.mockserver.url('/general_with'), self.parse_general, cb_kwargs=data)
        yield Request(self.mockserver.url('/general_without'), self.parse_general)
        yield Request(self.mockserver.url('/no_kwargs'), self.parse_no_kwargs)
        yield Request(self.mockserver.url('/default'), self.parse_default, cb_kwargs=data)
        yield Request(self.mockserver.url('/takes_less'), self.parse_takes_less, cb_kwargs=data)
        yield Request(self.mockserver.url('/takes_more'), self.parse_takes_more, cb_kwargs=data)
        yield Request(self.mockserver.url('/downloader_mw'), self.parse_downloader_mw)
        yield Request(self.mockserver.url('/spider_mw'), self.parse_spider_mw)

    def parse_first(self, response, key, number):
        self.checks.append(key == 'value')
        self.checks.append(number == 123)
        self.crawler.stats.inc_value('boolean_checks', 2)
        yield response.follow(
            self.mockserver.url('/two'),
            self.parse_second,
            cb_kwargs={'new_key': 'new_value'})

    def parse_second(self, response, new_key):
        self.checks.append(new_key == 'new_value')
        self.crawler.stats.inc_value('boolean_checks')

    def parse_general(self, response, **kwargs):
        if response.url.endswith('/general_with'):
            self.checks.append(kwargs['key'] == 'value')
            self.checks.append(kwargs['number'] == 123)
            self.checks.append(kwargs['callback'] == 'some_callback')
            self.crawler.stats.inc_value('boolean_checks', 3)
        elif response.url.endswith('/general_without'):
            self.checks.append(kwargs == {})
            self.crawler.stats.inc_value('boolean_checks')

    def parse_no_kwargs(self, response):
        self.checks.append(response.url.endswith('/no_kwargs'))
        self.crawler.stats.inc_value('boolean_checks')

    def parse_default(self, response, key, number=None, default=99):
        self.checks.append(response.url.endswith('/default'))
        self.checks.append(key == 'value')
        self.checks.append(number == 123)
        self.checks.append(default == 99)
        self.crawler.stats.inc_value('boolean_checks', 4)

    def parse_takes_less(self, response, key, callback):
        """
        Should raise
        TypeError: parse_takes_less() got an unexpected keyword argument 'number'
        """

    def parse_takes_more(self, response, key, number, callback, other):
        """
        Should raise
        TypeError: parse_takes_more() missing 1 required positional argument: 'other'
        """

    def parse_downloader_mw(self, response, from_process_request, from_process_response):
        self.checks.append(bool(from_process_request))
        self.checks.append(bool(from_process_response))
        self.crawler.stats.inc_value('boolean_checks', 2)

    def parse_spider_mw(self, response, from_process_spider_input, from_process_start_requests):
        self.checks.append(bool(from_process_spider_input))
        self.checks.append(bool(from_process_start_requests))
        self.crawler.stats.inc_value('boolean_checks', 2)
        return Request(self.mockserver.url('/spider_mw_2'), self.parse_spider_mw_2)

    def parse_spider_mw_2(self, response, from_process_spider_output):
        self.checks.append(bool(from_process_spider_output))
        self.crawler.stats.inc_value('boolean_checks', 1)


class CallbackKeywordArgumentsTestCase(TestCase):

    maxDiff = None

    def setUp(self):
        self.mockserver = MockServer()
        self.mockserver.__enter__()
        self.runner = CrawlerRunner()

    def tearDown(self):
        self.mockserver.__exit__(None, None, None)

    @defer.inlineCallbacks
    def test_callback_kwargs(self):
        crawler = self.runner.create_crawler(KeywordArgumentsSpider)
        with LogCapture() as log:
            yield crawler.crawl(mockserver=self.mockserver)
        self.assertTrue(all(crawler.spider.checks))
        self.assertEqual(len(crawler.spider.checks), crawler.stats.get_value('boolean_checks'))
        # check exceptions for argument mismatch
        exceptions = {}
        for line in log.records:
            for key in ('takes_less', 'takes_more'):
                if key in line.getMessage():
                    exceptions[key] = line
        self.assertEqual(exceptions['takes_less'].exc_info[0], TypeError)
        self.assertTrue(
            str(exceptions['takes_less'].exc_info[1]).endswith(
                "parse_takes_less() got an unexpected keyword argument 'number'"
            ),
            msg="Exception message: " + str(exceptions['takes_less'].exc_info[1]),
        )
        self.assertEqual(exceptions['takes_more'].exc_info[0], TypeError)
        self.assertTrue(
            str(exceptions['takes_more'].exc_info[1]).endswith(
                "parse_takes_more() missing 1 required positional argument: 'other'"
            ),
            msg="Exception message: " + str(exceptions['takes_more'].exc_info[1]),
        )



import unittest
import copy

from scrapy.http import Headers


class HeadersTest(unittest.TestCase):

    def assertSortedEqual(self, first, second, msg=None):
        return self.assertEqual(sorted(first), sorted(second), msg)

    def test_basics(self):
        h = Headers({'Content-Type': 'text/html', 'Content-Length': 1234})
        assert h['Content-Type']
        assert h['Content-Length']

        self.assertRaises(KeyError, h.__getitem__, 'Accept')
        self.assertEqual(h.get('Accept'), None)
        self.assertEqual(h.getlist('Accept'), [])

        self.assertEqual(h.get('Accept', '*/*'), b'*/*')
        self.assertEqual(h.getlist('Accept', '*/*'), [b'*/*'])
        self.assertEqual(h.getlist('Accept', ['text/html', 'images/jpeg']),
                         [b'text/html', b'images/jpeg'])

    def test_single_value(self):
        h = Headers()
        h['Content-Type'] = 'text/html'
        self.assertEqual(h['Content-Type'], b'text/html')
        self.assertEqual(h.get('Content-Type'), b'text/html')
        self.assertEqual(h.getlist('Content-Type'), [b'text/html'])

    def test_multivalue(self):
        h = Headers()
        h['X-Forwarded-For'] = hlist = ['ip1', 'ip2']
        self.assertEqual(h['X-Forwarded-For'], b'ip2')
        self.assertEqual(h.get('X-Forwarded-For'), b'ip2')
        self.assertEqual(h.getlist('X-Forwarded-For'), [b'ip1', b'ip2'])
        assert h.getlist('X-Forwarded-For') is not hlist

    def test_encode_utf8(self):
        h = Headers({'key': '\xa3'}, encoding='utf-8')
        key, val = dict(h).popitem()
        assert isinstance(key, bytes), key
        assert isinstance(val[0], bytes), val[0]
        self.assertEqual(val[0], b'\xc2\xa3')

    def test_encode_latin1(self):
        h = Headers({'key': '\xa3'}, encoding='latin1')
        key, val = dict(h).popitem()
        self.assertEqual(val[0], b'\xa3')

    def test_encode_multiple(self):
        h = Headers({'key': ['\xa3']}, encoding='utf-8')
        key, val = dict(h).popitem()
        self.assertEqual(val[0], b'\xc2\xa3')

    def test_delete_and_contains(self):
        h = Headers()
        h['Content-Type'] = 'text/html'
        assert 'Content-Type' in h
        del h['Content-Type']
        assert 'Content-Type' not in h

    def test_setdefault(self):
        h = Headers()
        hlist = ['ip1', 'ip2']
        olist = h.setdefault('X-Forwarded-For', hlist)
        assert h.getlist('X-Forwarded-For') is not hlist
        assert h.getlist('X-Forwarded-For') is olist

        h = Headers()
        olist = h.setdefault('X-Forwarded-For', 'ip1')
        self.assertEqual(h.getlist('X-Forwarded-For'), [b'ip1'])
        assert h.getlist('X-Forwarded-For') is olist

    def test_iterables(self):
        idict = {'Content-Type': 'text/html', 'X-Forwarded-For': ['ip1', 'ip2']}

        h = Headers(idict)
        self.assertDictEqual(dict(h),
                             {b'Content-Type': [b'text/html'],
                              b'X-Forwarded-For': [b'ip1', b'ip2']})
        self.assertSortedEqual(h.keys(),
                               [b'X-Forwarded-For', b'Content-Type'])
        self.assertSortedEqual(h.items(),
                               [(b'X-Forwarded-For', [b'ip1', b'ip2']),
                                (b'Content-Type', [b'text/html'])])
        self.assertSortedEqual(h.values(), [b'ip2', b'text/html'])

    def test_update(self):
        h = Headers()
        h.update({'Content-Type': 'text/html',
                  'X-Forwarded-For': ['ip1', 'ip2']})
        self.assertEqual(h.getlist('Content-Type'), [b'text/html'])
        self.assertEqual(h.getlist('X-Forwarded-For'), [b'ip1', b'ip2'])

    def test_copy(self):
        h1 = Headers({'header1': ['value1', 'value2']})
        h2 = copy.copy(h1)
        self.assertEqual(h1, h2)
        self.assertEqual(h1.getlist('header1'), h2.getlist('header1'))
        assert h1.getlist('header1') is not h2.getlist('header1')
        assert isinstance(h2, Headers)

    def test_appendlist(self):
        h1 = Headers({'header1': 'value1'})
        h1.appendlist('header1', 'value3')
        self.assertEqual(h1.getlist('header1'), [b'value1', b'value3'])

        h1 = Headers()
        h1.appendlist('header1', 'value1')
        h1.appendlist('header1', 'value3')
        self.assertEqual(h1.getlist('header1'), [b'value1', b'value3'])

    def test_setlist(self):
        h1 = Headers({'header1': 'value1'})
        self.assertEqual(h1.getlist('header1'), [b'value1'])
        h1.setlist('header1', [b'value2', b'value3'])
        self.assertEqual(h1.getlist('header1'), [b'value2', b'value3'])

    def test_setlistdefault(self):
        h1 = Headers({'header1': 'value1'})
        h1.setlistdefault('header1', ['value2', 'value3'])
        h1.setlistdefault('header2', ['value2', 'value3'])
        self.assertEqual(h1.getlist('header1'), [b'value1'])
        self.assertEqual(h1.getlist('header2'), [b'value2', b'value3'])

    def test_none_value(self):
        h1 = Headers()
        h1['foo'] = 'bar'
        h1['foo'] = None
        h1.setdefault('foo', 'bar')
        self.assertEqual(h1.get('foo'), None)
        self.assertEqual(h1.getlist('foo'), [])

    def test_int_value(self):
        h1 = Headers({'hey': 5})
        h1['foo'] = 1
        h1.setdefault('bar', 2)
        h1.setlist('buz', [1, 'dos', 3])
        self.assertEqual(h1.getlist('foo'), [b'1'])
        self.assertEqual(h1.getlist('bar'), [b'2'])
        self.assertEqual(h1.getlist('buz'), [b'1', b'dos', b'3'])
        self.assertEqual(h1.getlist('hey'), [b'5'])

    def test_invalid_value(self):
        self.assertRaisesRegex(TypeError, 'Unsupported value type',
                               Headers, {'foo': object()})
        self.assertRaisesRegex(TypeError, 'Unsupported value type',
                               Headers().__setitem__, 'foo', object())
        self.assertRaisesRegex(TypeError, 'Unsupported value type',
                               Headers().setdefault, 'foo', object())
        self.assertRaisesRegex(TypeError, 'Unsupported value type',
                               Headers().setlist, 'foo', [object()])



# 爬虫中间件 ItemHeadersMiddleware 结合 oss下载中间键使用 添加请求头参数
SPIDER_MIDDLEWARES = {
    "Scrapy_Prj.middlewares.spidermiddlewares.itemheaders.ItemHeadersMiddleware":300
}

# 下载中间件配置参数
DOWNLOADER_MIDDLEWARES = {
    "gerapy_pyppeteer.downloadermiddlewares.PyppeteerMiddleware":540,
    'Scrapy_Prj.middlewares.downloadermiddlewares.randuamiddleware.RandomUserAgent': 543,
    'Scrapy_Prj.middlewares.downloadermiddlewares.proxymiddleware.ProxyMiddleware': 888,
}

# 阿里云文件管道以及存储管道设置参数
ITEM_PIPELINES = {
    'Scrapy_Prj.pipelines.ossimgpipelines.OssImagesPipeline': 200,
    'Scrapy_Prj.pipelines.ossimgpipelines.OssImagesPipelineBak': 201,
    'Scrapy_Prj.pipelines.ossfilepipelines.OssFilesPipeline': 300,
    'Scrapy_Prj.pipelines.ossfilepipelines.OssFilesPipelineBak': 301,
    'Scrapy_Prj.pipelines.bloomerfiler.BloomFilterPipeline': 399,
    'Scrapy_Prj.pipelines.mysqlpipelines.MySQLPipeline': 400,
    'Scrapy_Prj.pipelines.mongodbPipelines.MongodbPipeline': 450,
    'Scrapy_Prj.pipelines.rabbitmqpipelines.RabbitMQPipeline': 500,
    'Scrapy_Prj.pipelines.excelpipelines.ExcelFactoryPipeline': 550,
    'Scrapy_Prj.pipelines.excelpipelines.ExcelDataplatformPipeline': 550,
    'Scrapy_Prj.pipelines.influxdbpipelines.SpiderStatLogging': 600,
}

# 拓展中间件
EXTENSIONS = {
    'scrapy.extensions.closespider.CloseSpider': 500, # CLOSESPIDER_TIMEOUT/CLOSESPIDER_ERRORCOUNT/CLOSESPIDER_ITEMCOUNT设置运行时间停止、错误个数停止或者item数量停止等
}


"""
tests: this package contains all Scrapy unittests

see https://docs.scrapy.org/en/latest/contributing.html#running-tests
"""

import os

# ignore system-wide proxies for tests
# which would send requests to a totally unsuspecting server
# (e.g. because urllib does not fully understand the proxy spec)
os.environ['http_proxy'] = ''
os.environ['https_proxy'] = ''
os.environ['ftp_proxy'] = ''

# Absolutize paths to coverage config and output file because tests that
# spawn subprocesses also changes current working directory.
_sourceroot = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
if 'COV_CORE_CONFIG' in os.environ:
    os.environ['COVERAGE_FILE'] = os.path.join(_sourceroot, '.coverage')
    os.environ['COV_CORE_CONFIG'] = os.path.join(_sourceroot,
                                                 os.environ['COV_CORE_CONFIG'])

tests_datadir = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                             'sample_data')


def get_testdata(*paths):
    """Return test data"""
    path = os.path.join(tests_datadir, *paths)
    with open(path, 'rb') as f:
        return f.read()



import json
import os
import random
import re
import shutil
import string
from ipaddress import IPv4Address
from unittest import mock, skipIf
from urllib.parse import urlencode

from twisted.internet import reactor
from twisted.internet.defer import CancelledError, Deferred, DeferredList, inlineCallbacks
from twisted.internet.endpoints import SSL4ClientEndpoint, SSL4ServerEndpoint
from twisted.internet.error import TimeoutError
from twisted.internet.ssl import optionsForClientTLS, PrivateCertificate, Certificate
from twisted.python.failure import Failure
from twisted.trial.unittest import TestCase
from twisted.web.client import ResponseFailed, URI
from twisted.web.http import H2_ENABLED, Request as TxRequest
from twisted.web.server import Site, NOT_DONE_YET
from twisted.web.static import File

from scrapy.http import Request, Response, JsonRequest
from scrapy.settings import Settings
from scrapy.spiders import Spider
from tests.mockserver import ssl_context_factory, LeafResource, Status


def generate_random_string(size):
    return ''.join(random.choices(
        string.ascii_uppercase + string.digits,
        k=size
    ))


def make_html_body(val):
    response = f'''<html>
<h1>Hello from HTTP2<h1>
<p>{val}</p>
</html>'''
    return bytes(response, 'utf-8')


class DummySpider(Spider):
    name = 'dummy'
    start_urls: list = []

    def parse(self, response):
        print(response)


class Data:
    SMALL_SIZE = 1024  # 1 KB
    LARGE_SIZE = 1024 ** 2  # 1 MB

    STR_SMALL = generate_random_string(SMALL_SIZE)
    STR_LARGE = generate_random_string(LARGE_SIZE)

    EXTRA_SMALL = generate_random_string(1024 * 15)
    EXTRA_LARGE = generate_random_string((1024 ** 2) * 15)

    HTML_SMALL = make_html_body(STR_SMALL)
    HTML_LARGE = make_html_body(STR_LARGE)

    JSON_SMALL = {'data': STR_SMALL}
    JSON_LARGE = {'data': STR_LARGE}

    DATALOSS = b'Dataloss Content'
    NO_CONTENT_LENGTH = b'This response do not have any content-length header'


class GetDataHtmlSmall(LeafResource):
    def render_GET(self, request: TxRequest):
        request.setHeader('Content-Type', 'text/html; charset=UTF-8')
        return Data.HTML_SMALL


class GetDataHtmlLarge(LeafResource):
    def render_GET(self, request: TxRequest):
        request.setHeader('Content-Type', 'text/html; charset=UTF-8')
        return Data.HTML_LARGE


class PostDataJsonMixin:
    @staticmethod
    def make_response(request: TxRequest, extra_data: str):
        response = {
            'request-headers': {},
            'request-body': json.loads(request.content.read()),
            'extra-data': extra_data
        }
        for k, v in request.requestHeaders.getAllRawHeaders():
            response['request-headers'][str(k, 'utf-8')] = str(v[0], 'utf-8')

        response_bytes = bytes(json.dumps(response), 'utf-8')
        request.setHeader('Content-Type', 'application/json; charset=UTF-8')
        request.setHeader('Content-Encoding', 'UTF-8')
        return response_bytes


class PostDataJsonSmall(LeafResource, PostDataJsonMixin):
    def render_POST(self, request: TxRequest):
        return self.make_response(request, Data.EXTRA_SMALL)


class PostDataJsonLarge(LeafResource, PostDataJsonMixin):
    def render_POST(self, request: TxRequest):
        return self.make_response(request, Data.EXTRA_LARGE)


class Dataloss(LeafResource):

    def render_GET(self, request: TxRequest):
        request.setHeader(b"Content-Length", b"1024")
        self.deferRequest(request, 0, self._delayed_render, request)
        return NOT_DONE_YET

    @staticmethod
    def _delayed_render(request: TxRequest):
        request.write(Data.DATALOSS)
        request.finish()


class NoContentLengthHeader(LeafResource):
    def render_GET(self, request: TxRequest):
        request.requestHeaders.removeHeader('Content-Length')
        self.deferRequest(request, 0, self._delayed_render, request)
        return NOT_DONE_YET

    @staticmethod
    def _delayed_render(request: TxRequest):
        request.write(Data.NO_CONTENT_LENGTH)
        request.finish()


class TimeoutResponse(LeafResource):
    def render_GET(self, request: TxRequest):
        return NOT_DONE_YET


class QueryParams(LeafResource):
    def render_GET(self, request: TxRequest):
        request.setHeader('Content-Type', 'application/json; charset=UTF-8')
        request.setHeader('Content-Encoding', 'UTF-8')

        query_params = {}
        for k, v in request.args.items():
            query_params[str(k, 'utf-8')] = str(v[0], 'utf-8')

        return bytes(json.dumps(query_params), 'utf-8')


class RequestHeaders(LeafResource):
    """Sends all the headers received as a response"""

    def render_GET(self, request: TxRequest):
        request.setHeader('Content-Type', 'application/json; charset=UTF-8')
        request.setHeader('Content-Encoding', 'UTF-8')
        headers = {}
        for k, v in request.requestHeaders.getAllRawHeaders():
            headers[str(k, 'utf-8')] = str(v[0], 'utf-8')

        return bytes(json.dumps(headers), 'utf-8')


def get_client_certificate(key_file, certificate_file) -> PrivateCertificate:
    with open(key_file, 'r') as key, open(certificate_file, 'r') as certificate:
        pem = ''.join(key.readlines()) + ''.join(certificate.readlines())

    return PrivateCertificate.loadPEM(pem)


@skipIf(not H2_ENABLED, "HTTP/2 support in Twisted is not enabled")
class Https2ClientProtocolTestCase(TestCase):
    scheme = 'https'
    key_file = os.path.join(os.path.dirname(__file__), 'keys', 'localhost.key')
    certificate_file = os.path.join(os.path.dirname(__file__), 'keys', 'localhost.crt')

    def _init_resource(self):
        self.temp_directory = self.mktemp()
        os.mkdir(self.temp_directory)
        r = File(self.temp_directory)
        r.putChild(b'get-data-html-small', GetDataHtmlSmall())
        r.putChild(b'get-data-html-large', GetDataHtmlLarge())

        r.putChild(b'post-data-json-small', PostDataJsonSmall())
        r.putChild(b'post-data-json-large', PostDataJsonLarge())

        r.putChild(b'dataloss', Dataloss())
        r.putChild(b'no-content-length-header', NoContentLengthHeader())
        r.putChild(b'status', Status())
        r.putChild(b'query-params', QueryParams())
        r.putChild(b'timeout', TimeoutResponse())
        r.putChild(b'request-headers', RequestHeaders())
        return r

    @inlineCallbacks
    def setUp(self):
        # Initialize resource tree
        root = self._init_resource()
        self.site = Site(root, timeout=None)

        # Start server for testing
        self.hostname = 'localhost'
        context_factory = ssl_context_factory(self.key_file, self.certificate_file)

        server_endpoint = SSL4ServerEndpoint(reactor, 0, context_factory, interface=self.hostname)
        self.server = yield server_endpoint.listen(self.site)
        self.port_number = self.server.getHost().port

        # Connect H2 client with server
        self.client_certificate = get_client_certificate(self.key_file, self.certificate_file)
        client_options = optionsForClientTLS(
            hostname=self.hostname,
            trustRoot=self.client_certificate,
            acceptableProtocols=[b'h2']
        )
        uri = URI.fromBytes(bytes(self.get_url('/'), 'utf-8'))

        self.conn_closed_deferred = Deferred()
        from scrapy.core.http2.protocol import H2ClientFactory
        h2_client_factory = H2ClientFactory(uri, Settings(), self.conn_closed_deferred)
        client_endpoint = SSL4ClientEndpoint(reactor, self.hostname, self.port_number, client_options)
        self.client = yield client_endpoint.connect(h2_client_factory)

    @inlineCallbacks
    def tearDown(self):
        if self.client.connected:
            yield self.client.transport.loseConnection()
            yield self.client.transport.abortConnection()
        yield self.server.stopListening()
        shutil.rmtree(self.temp_directory)
        self.conn_closed_deferred = None

    def get_url(self, path):
        """
        :param path: Should have / at the starting compulsorily if not empty
        :return: Complete url
        """
        assert len(path) > 0 and (path[0] == '/' or path[0] == '&')
        return f'{self.scheme}://{self.hostname}:{self.port_number}{path}'

    def make_request(self, request: Request) -> Deferred:
        return self.client.request(request, DummySpider())

    @staticmethod
    def _check_repeat(get_deferred, count):
        d_list = []
        for _ in range(count):
            d = get_deferred()
            d_list.append(d)

        return DeferredList(d_list, fireOnOneErrback=True)

    def _check_GET(
        self,
        request: Request,
        expected_body,
        expected_status
    ):
        def check_response(response: Response):
            self.assertEqual(response.status, expected_status)
            self.assertEqual(response.body, expected_body)
            self.assertEqual(response.request, request)

            content_length = int(response.headers.get('Content-Length'))
            self.assertEqual(len(response.body), content_length)

        d = self.make_request(request)
        d.addCallback(check_response)
        d.addErrback(self.fail)
        return d

    def test_GET_small_body(self):
        request = Request(self.get_url('/get-data-html-small'))
        return self._check_GET(request, Data.HTML_SMALL, 200)

    def test_GET_large_body(self):
        request = Request(self.get_url('/get-data-html-large'))
        return self._check_GET(request, Data.HTML_LARGE, 200)

    def _check_GET_x10(self, *args, **kwargs):
        def get_deferred():
            return self._check_GET(*args, **kwargs)

        return self._check_repeat(get_deferred, 10)

    def test_GET_small_body_x10(self):
        return self._check_GET_x10(
            Request(self.get_url('/get-data-html-small')),
            Data.HTML_SMALL,
            200
        )

    def test_GET_large_body_x10(self):
        return self._check_GET_x10(
            Request(self.get_url('/get-data-html-large')),
            Data.HTML_LARGE,
            200
        )

    def _check_POST_json(
        self,
        request: Request,
        expected_request_body,
        expected_extra_data,
        expected_status: int
    ):
        d = self.make_request(request)

        def assert_response(response: Response):
            self.assertEqual(response.status, expected_status)
            self.assertEqual(response.request, request)

            content_length = int(response.headers.get('Content-Length'))
            self.assertEqual(len(response.body), content_length)

            # Parse the body
            content_encoding = str(response.headers[b'Content-Encoding'], 'utf-8')
            body = json.loads(str(response.body, content_encoding))
            self.assertIn('request-body', body)
            self.assertIn('extra-data', body)
            self.assertIn('request-headers', body)

            request_body = body['request-body']
            self.assertEqual(request_body, expected_request_body)

            extra_data = body['extra-data']
            self.assertEqual(extra_data, expected_extra_data)

            # Check if headers were sent successfully
            request_headers = body['request-headers']
            for k, v in request.headers.items():
                k_str = str(k, 'utf-8')
                self.assertIn(k_str, request_headers)
                self.assertEqual(request_headers[k_str], str(v[0], 'utf-8'))

        d.addCallback(assert_response)
        d.addErrback(self.fail)
        return d

    def test_POST_small_json(self):
        request = JsonRequest(url=self.get_url('/post-data-json-small'), method='POST', data=Data.JSON_SMALL)
        return self._check_POST_json(
            request,
            Data.JSON_SMALL,
            Data.EXTRA_SMALL,
            200
        )

    def test_POST_large_json(self):
        request = JsonRequest(url=self.get_url('/post-data-json-large'), method='POST', data=Data.JSON_LARGE)
        return self._check_POST_json(
            request,
            Data.JSON_LARGE,
            Data.EXTRA_LARGE,
            200
        )

    def _check_POST_json_x10(self, *args, **kwargs):
        def get_deferred():
            return self._check_POST_json(*args, **kwargs)

        return self._check_repeat(get_deferred, 10)

    def test_POST_small_json_x10(self):
        request = JsonRequest(url=self.get_url('/post-data-json-small'), method='POST', data=Data.JSON_SMALL)
        return self._check_POST_json_x10(
            request,
            Data.JSON_SMALL,
            Data.EXTRA_SMALL,
            200
        )

    def test_POST_large_json_x10(self):
        request = JsonRequest(url=self.get_url('/post-data-json-large'), method='POST', data=Data.JSON_LARGE)
        return self._check_POST_json_x10(
            request,
            Data.JSON_LARGE,
            Data.EXTRA_LARGE,
            200
        )

    @inlineCallbacks
    def test_invalid_negotiated_protocol(self):
        with mock.patch("scrapy.core.http2.protocol.PROTOCOL_NAME", return_value=b"not-h2"):
            request = Request(url=self.get_url('/status?n=200'))
            with self.assertRaises(ResponseFailed):
                yield self.make_request(request)

    def test_cancel_request(self):
        request = Request(url=self.get_url('/get-data-html-large'))

        def assert_response(response: Response):
            self.assertEqual(response.status, 499)
            self.assertEqual(response.request, request)

        d = self.make_request(request)
        d.addCallback(assert_response)
        d.addErrback(self.fail)
        d.cancel()

        return d

    def test_download_maxsize_exceeded(self):
        request = Request(url=self.get_url('/get-data-html-large'), meta={'download_maxsize': 1000})

        def assert_cancelled_error(failure):
            self.assertIsInstance(failure.value, CancelledError)
            error_pattern = re.compile(
                rf'Cancelling download of {request.url}: received response '
                rf'size \(\d*\) larger than download max size \(1000\)'
            )
            self.assertEqual(len(re.findall(error_pattern, str(failure.value))), 1)

        d = self.make_request(request)
        d.addCallback(self.fail)
        d.addErrback(assert_cancelled_error)
        return d

    def test_received_dataloss_response(self):
        """In case when value of Header Content-Length != len(Received Data)
        ProtocolError is raised"""
        request = Request(url=self.get_url('/dataloss'))

        def assert_failure(failure: Failure):
            self.assertTrue(len(failure.value.reasons) > 0)
            from h2.exceptions import InvalidBodyLengthError
            self.assertTrue(any(
                isinstance(error, InvalidBodyLengthError)
                for error in failure.value.reasons
            ))

        d = self.make_request(request)
        d.addCallback(self.fail)
        d.addErrback(assert_failure)
        return d

    def test_missing_content_length_header(self):
        request = Request(url=self.get_url('/no-content-length-header'))

        def assert_content_length(response: Response):
            self.assertEqual(response.status, 200)
            self.assertEqual(response.body, Data.NO_CONTENT_LENGTH)
            self.assertEqual(response.request, request)
            self.assertNotIn('Content-Length', response.headers)

        d = self.make_request(request)
        d.addCallback(assert_content_length)
        d.addErrback(self.fail)
        return d

    @inlineCallbacks
    def _check_log_warnsize(
        self,
        request,
        warn_pattern,
        expected_body
    ):
        with self.assertLogs('scrapy.core.http2.stream', level='WARNING') as cm:
            response = yield self.make_request(request)
            self.assertEqual(response.status, 200)
            self.assertEqual(response.request, request)
            self.assertEqual(response.body, expected_body)

            # Check the warning is raised only once for this request
            self.assertEqual(sum(
                len(re.findall(warn_pattern, log))
                for log in cm.output
            ), 1)

    @inlineCallbacks
    def test_log_expected_warnsize(self):
        request = Request(url=self.get_url('/get-data-html-large'), meta={'download_warnsize': 1000})
        warn_pattern = re.compile(
            rf'Expected response size \(\d*\) larger than '
            rf'download warn size \(1000\) in request {request}'
        )

        yield self._check_log_warnsize(request, warn_pattern, Data.HTML_LARGE)

    @inlineCallbacks
    def test_log_received_warnsize(self):
        request = Request(url=self.get_url('/no-content-length-header'), meta={'download_warnsize': 10})
        warn_pattern = re.compile(
            rf'Received more \(\d*\) bytes than download '
            rf'warn size \(10\) in request {request}'
        )

        yield self._check_log_warnsize(request, warn_pattern, Data.NO_CONTENT_LENGTH)

    def test_max_concurrent_streams(self):
        """Send 500 requests at one to check if we can handle
        very large number of request.
        """

        def get_deferred():
            return self._check_GET(
                Request(self.get_url('/get-data-html-small')),
                Data.HTML_SMALL,
                200
            )

        return self._check_repeat(get_deferred, 500)

    def test_inactive_stream(self):
        """Here we send 110 requests considering the MAX_CONCURRENT_STREAMS
        by default is 100. After sending the first 100 requests we close the
        connection."""
        d_list = []

        def assert_inactive_stream(failure):
            self.assertIsNotNone(failure.check(ResponseFailed))
            from scrapy.core.http2.stream import InactiveStreamClosed
            self.assertTrue(any(
                isinstance(e, InactiveStreamClosed)
                for e in failure.value.reasons
            ))

        # Send 100 request (we do not check the result)
        for _ in range(100):
            d = self.make_request(Request(self.get_url('/get-data-html-small')))
            d.addBoth(lambda _: None)
            d_list.append(d)

        # Now send 10 extra request and save the response deferred in a list
        for _ in range(10):
            d = self.make_request(Request(self.get_url('/get-data-html-small')))
            d.addCallback(self.fail)
            d.addErrback(assert_inactive_stream)
            d_list.append(d)

        # Close the connection now to fire all the extra 10 requests errback
        # with InactiveStreamClosed
        self.client.transport.loseConnection()

        return DeferredList(d_list, consumeErrors=True, fireOnOneErrback=True)

    def test_invalid_request_type(self):
        with self.assertRaises(TypeError):
            self.make_request('https://InvalidDataTypePassed.com')

    def test_query_parameters(self):
        params = {
            'a': generate_random_string(20),
            'b': generate_random_string(20),
            'c': generate_random_string(20),
            'd': generate_random_string(20)
        }
        request = Request(self.get_url(f'/query-params?{urlencode(params)}'))

        def assert_query_params(response: Response):
            content_encoding = str(response.headers[b'Content-Encoding'], 'utf-8')
            data = json.loads(str(response.body, content_encoding))
            self.assertEqual(data, params)

        d = self.make_request(request)
        d.addCallback(assert_query_params)
        d.addErrback(self.fail)

        return d

    def test_status_codes(self):
        def assert_response_status(response: Response, expected_status: int):
            self.assertEqual(response.status, expected_status)

        d_list = []
        for status in [200, 404]:
            request = Request(self.get_url(f'/status?n={status}'))
            d = self.make_request(request)
            d.addCallback(assert_response_status, status)
            d.addErrback(self.fail)
            d_list.append(d)

        return DeferredList(d_list, fireOnOneErrback=True)

    def test_response_has_correct_certificate_ip_address(self):
        request = Request(self.get_url('/status?n=200'))

        def assert_metadata(response: Response):
            self.assertEqual(response.request, request)
            self.assertIsInstance(response.certificate, Certificate)
            self.assertIsNotNone(response.certificate.original)
            self.assertEqual(response.certificate.getIssuer(), self.client_certificate.getIssuer())
            self.assertTrue(response.certificate.getPublicKey().matches(self.client_certificate.getPublicKey()))

            self.assertIsInstance(response.ip_address, IPv4Address)
            self.assertEqual(str(response.ip_address), '127.0.0.1')

        d = self.make_request(request)
        d.addCallback(assert_metadata)
        d.addErrback(self.fail)

        return d

    def _check_invalid_netloc(self, url):
        request = Request(url)

        def assert_invalid_hostname(failure: Failure):
            from scrapy.core.http2.stream import InvalidHostname
            self.assertIsNotNone(failure.check(InvalidHostname))
            error_msg = str(failure.value)
            self.assertIn('localhost', error_msg)
            self.assertIn('127.0.0.1', error_msg)
            self.assertIn(str(request), error_msg)

        d = self.make_request(request)
        d.addCallback(self.fail)
        d.addErrback(assert_invalid_hostname)
        return d

    def test_invalid_hostname(self):
        return self._check_invalid_netloc('https://notlocalhost.notlocalhostdomain')

    def test_invalid_host_port(self):
        port = self.port_number + 1
        return self._check_invalid_netloc(f'https://127.0.0.1:{port}')

    def test_connection_stays_with_invalid_requests(self):
        d_list = [
            self.test_invalid_hostname(),
            self.test_invalid_host_port(),
            self.test_GET_small_body(),
            self.test_POST_small_json()
        ]

        return DeferredList(d_list, fireOnOneErrback=True)

    def test_connection_timeout(self):
        request = Request(self.get_url('/timeout'))
        d = self.make_request(request)

        # Update the timer to 1s to test connection timeout
        self.client.setTimeout(1)

        def assert_timeout_error(failure: Failure):
            for err in failure.value.reasons:
                from scrapy.core.http2.protocol import H2ClientProtocol
                if isinstance(err, TimeoutError):
                    self.assertIn(f"Connection was IDLE for more than {H2ClientProtocol.IDLE_TIMEOUT}s", str(err))
                    break
            else:
                self.fail()

        d.addCallback(self.fail)
        d.addErrback(assert_timeout_error)
        return d

    def test_request_headers_received(self):
        request = Request(self.get_url('/request-headers'), headers={
            'header-1': 'header value 1',
            'header-2': 'header value 2'
        })
        d = self.make_request(request)

        def assert_request_headers(response: Response):
            self.assertEqual(response.status, 200)
            self.assertEqual(response.request, request)

            response_headers = json.loads(str(response.body, 'utf-8'))
            self.assertIsInstance(response_headers, dict)
            for k, v in request.headers.items():
                k, v = str(k, 'utf-8'), str(v[0], 'utf-8')
                self.assertIn(k, response_headers)
                self.assertEqual(v, response_headers[k])

        d.addErrback(self.fail)
        d.addCallback(assert_request_headers)
        return d



"""
Mail sending helpers

See documentation in docs/topics/email.rst
"""
import logging
from email import encoders as Encoders
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.nonmultipart import MIMENonMultipart
from email.mime.text import MIMEText
from email.utils import COMMASPACE, formatdate
from io import BytesIO

from twisted.internet import defer, ssl

from scrapy.utils.misc import arg_to_iter
from scrapy.utils.python import to_bytes


logger = logging.getLogger(__name__)


def _to_bytes_or_none(text):
    if text is None:
        return None
    return to_bytes(text)


class MailSender:
    def __init__(
        self, smtphost='localhost', mailfrom='scrapy@localhost', smtpuser=None,
        smtppass=None, smtpport=25, smtptls=False, smtpssl=False, debug=False
    ):
        self.smtphost = smtphost
        self.smtpport = smtpport
        self.smtpuser = _to_bytes_or_none(smtpuser)
        self.smtppass = _to_bytes_or_none(smtppass)
        self.smtptls = smtptls
        self.smtpssl = smtpssl
        self.mailfrom = mailfrom
        self.debug = debug

    @classmethod
    def from_settings(cls, settings):
        return cls(
            smtphost=settings['MAIL_HOST'],
            mailfrom=settings['MAIL_FROM'],
            smtpuser=settings['MAIL_USER'],
            smtppass=settings['MAIL_PASS'],
            smtpport=settings.getint('MAIL_PORT'),
            smtptls=settings.getbool('MAIL_TLS'),
            smtpssl=settings.getbool('MAIL_SSL'),
        )

    def send(self, to, subject, body, cc=None, attachs=(), mimetype='text/plain', charset=None, _callback=None):
        from twisted.internet import reactor
        if attachs:
            msg = MIMEMultipart()
        else:
            msg = MIMENonMultipart(*mimetype.split('/', 1))

        to = list(arg_to_iter(to))
        cc = list(arg_to_iter(cc))

        msg['From'] = self.mailfrom
        msg['To'] = COMMASPACE.join(to)
        msg['Date'] = formatdate(localtime=True)
        msg['Subject'] = subject
        rcpts = to[:]
        if cc:
            rcpts.extend(cc)
            msg['Cc'] = COMMASPACE.join(cc)

        if charset:
            msg.set_charset(charset)

        if attachs:
            msg.attach(MIMEText(body, 'plain', charset or 'us-ascii'))
            for attach_name, mimetype, f in attachs:
                part = MIMEBase(*mimetype.split('/'))
                part.set_payload(f.read())
                Encoders.encode_base64(part)
                part.add_header('Content-Disposition', 'attachment', filename=attach_name)
                msg.attach(part)
        else:
            msg.set_payload(body)

        if _callback:
            _callback(to=to, subject=subject, body=body, cc=cc, attach=attachs, msg=msg)

        if self.debug:
            logger.debug('Debug mail sent OK: To=%(mailto)s Cc=%(mailcc)s '
                         'Subject="%(mailsubject)s" Attachs=%(mailattachs)d',
                         {'mailto': to, 'mailcc': cc, 'mailsubject': subject,
                          'mailattachs': len(attachs)})
            return

        dfd = self._sendmail(rcpts, msg.as_string().encode(charset or 'utf-8'))
        dfd.addCallbacks(
            callback=self._sent_ok,
            errback=self._sent_failed,
            callbackArgs=[to, cc, subject, len(attachs)],
            errbackArgs=[to, cc, subject, len(attachs)],
        )
        reactor.addSystemEventTrigger('before', 'shutdown', lambda: dfd)
        return dfd

    def _sent_ok(self, result, to, cc, subject, nattachs):
        logger.info('Mail sent OK: To=%(mailto)s Cc=%(mailcc)s '
                    'Subject="%(mailsubject)s" Attachs=%(mailattachs)d',
                    {'mailto': to, 'mailcc': cc, 'mailsubject': subject,
                     'mailattachs': nattachs})

    def _sent_failed(self, failure, to, cc, subject, nattachs):
        errstr = str(failure.value)
        logger.error('Unable to send mail: To=%(mailto)s Cc=%(mailcc)s '
                     'Subject="%(mailsubject)s" Attachs=%(mailattachs)d'
                     '- %(mailerr)s',
                     {'mailto': to, 'mailcc': cc, 'mailsubject': subject,
                      'mailattachs': nattachs, 'mailerr': errstr})

    def _sendmail(self, to_addrs, msg):
        # Import twisted.mail here because it is not available in python3
        from twisted.internet import reactor
        from twisted.mail.smtp import ESMTPSenderFactory
        msg = BytesIO(msg)
        d = defer.Deferred()
        factory = ESMTPSenderFactory(
            self.smtpuser, self.smtppass, self.mailfrom, to_addrs, msg, d,
            heloFallback=True, requireAuthentication=False, requireTransportSecurity=self.smtptls,
        )
        factory.noisy = False

        if self.smtpssl:
            reactor.connectSSL(self.smtphost, self.smtpport, factory, ssl.ClientContextFactory())
        else:
            reactor.connectTCP(self.smtphost, self.smtpport, factory)

        return d



import sys
import unittest
import warnings
from contextlib import suppress

from scrapy import Spider, Request
from scrapy.exceptions import ScrapyDeprecationWarning
from scrapy.http import FormRequest, JsonRequest
from scrapy.utils.request import request_from_dict


class CustomRequest(Request):
    pass


class RequestSerializationTest(unittest.TestCase):

    def setUp(self):
        self.spider = TestSpider()

    def test_basic(self):
        r = Request("http://www.example.com")
        self._assert_serializes_ok(r)

    def test_all_attributes(self):
        r = Request(
            url="http://www.example.com",
            callback=self.spider.parse_item,
            errback=self.spider.handle_error,
            method="POST",
            body=b"some body",
            headers={'content-encoding': 'text/html; charset=latin-1'},
            cookies={'currency': 'руб'},
            encoding='latin-1',
            priority=20,
            meta={'a': 'b'},
            cb_kwargs={'k': 'v'},
            flags=['testFlag'],
        )
        self._assert_serializes_ok(r, spider=self.spider)

    def test_latin1_body(self):
        r = Request("http://www.example.com", body=b"\xa3")
        self._assert_serializes_ok(r)

    def test_utf8_body(self):
        r = Request("http://www.example.com", body=b"\xc2\xa3")
        self._assert_serializes_ok(r)

    def _assert_serializes_ok(self, request, spider=None):
        d = request.to_dict(spider=spider)
        request2 = request_from_dict(d, spider=spider)
        self._assert_same_request(request, request2)

    def _assert_same_request(self, r1, r2):
        self.assertEqual(r1.__class__, r2.__class__)
        self.assertEqual(r1.url, r2.url)
        self.assertEqual(r1.callback, r2.callback)
        self.assertEqual(r1.errback, r2.errback)
        self.assertEqual(r1.method, r2.method)
        self.assertEqual(r1.body, r2.body)
        self.assertEqual(r1.headers, r2.headers)
        self.assertEqual(r1.cookies, r2.cookies)
        self.assertEqual(r1.meta, r2.meta)
        self.assertEqual(r1.cb_kwargs, r2.cb_kwargs)
        self.assertEqual(r1.encoding, r2.encoding)
        self.assertEqual(r1._encoding, r2._encoding)
        self.assertEqual(r1.priority, r2.priority)
        self.assertEqual(r1.dont_filter, r2.dont_filter)
        self.assertEqual(r1.flags, r2.flags)
        if isinstance(r1, JsonRequest):
            self.assertEqual(r1.dumps_kwargs, r2.dumps_kwargs)

    def test_request_class(self):
        r1 = FormRequest("http://www.example.com")
        self._assert_serializes_ok(r1, spider=self.spider)
        r2 = CustomRequest("http://www.example.com")
        self._assert_serializes_ok(r2, spider=self.spider)
        r3 = JsonRequest("http://www.example.com", dumps_kwargs={"indent": 4})
        self._assert_serializes_ok(r3, spider=self.spider)

    def test_callback_serialization(self):
        r = Request("http://www.example.com", callback=self.spider.parse_item,
                    errback=self.spider.handle_error)
        self._assert_serializes_ok(r, spider=self.spider)

    def test_reference_callback_serialization(self):
        r = Request("http://www.example.com",
                    callback=self.spider.parse_item_reference,
                    errback=self.spider.handle_error_reference)
        self._assert_serializes_ok(r, spider=self.spider)
        request_dict = r.to_dict(spider=self.spider)
        self.assertEqual(request_dict['callback'], 'parse_item_reference')
        self.assertEqual(request_dict['errback'], 'handle_error_reference')

    def test_private_reference_callback_serialization(self):
        r = Request("http://www.example.com",
                    callback=self.spider._TestSpider__parse_item_reference,
                    errback=self.spider._TestSpider__handle_error_reference)
        self._assert_serializes_ok(r, spider=self.spider)
        request_dict = r.to_dict(spider=self.spider)
        self.assertEqual(request_dict['callback'],
                         '_TestSpider__parse_item_reference')
        self.assertEqual(request_dict['errback'],
                         '_TestSpider__handle_error_reference')

    def test_private_callback_serialization(self):
        r = Request("http://www.example.com",
                    callback=self.spider._TestSpider__parse_item_private,
                    errback=self.spider.handle_error)
        self._assert_serializes_ok(r, spider=self.spider)

    def test_mixin_private_callback_serialization(self):
        r = Request("http://www.example.com",
                    callback=self.spider._TestSpiderMixin__mixin_callback,
                    errback=self.spider.handle_error)
        self._assert_serializes_ok(r, spider=self.spider)

    def test_delegated_callback_serialization(self):
        r = Request("http://www.example.com",
                    callback=self.spider.delegated_callback,
                    errback=self.spider.handle_error)
        self._assert_serializes_ok(r, spider=self.spider)

    def test_unserializable_callback1(self):
        r = Request("http://www.example.com", callback=lambda x: x)
        self.assertRaises(ValueError, r.to_dict, spider=self.spider)

    def test_unserializable_callback2(self):
        r = Request("http://www.example.com", callback=self.spider.parse_item)
        self.assertRaises(ValueError, r.to_dict, spider=None)

    def test_unserializable_callback3(self):
        """Parser method is removed or replaced dynamically."""

        class MySpider(Spider):
            name = 'my_spider'

            def parse(self, response):
                pass

        spider = MySpider()
        r = Request("http://www.example.com", callback=spider.parse)
        setattr(spider, 'parse', None)
        self.assertRaises(ValueError, r.to_dict, spider=spider)

    def test_callback_not_available(self):
        """Callback method is not available in the spider passed to from_dict"""
        spider = TestSpiderDelegation()
        r = Request("http://www.example.com", callback=spider.delegated_callback)
        d = r.to_dict(spider=spider)
        self.assertRaises(ValueError, request_from_dict, d, spider=Spider("foo"))


class DeprecatedMethodsRequestSerializationTest(RequestSerializationTest):
    def _assert_serializes_ok(self, request, spider=None):
        with warnings.catch_warnings(record=True) as caught:
            warnings.simplefilter("always")
            with suppress(KeyError):
                del sys.modules["scrapy.utils.reqser"]  # delete module to reset the deprecation warning

            from scrapy.utils.reqser import request_from_dict as _from_dict, request_to_dict as _to_dict

            request_copy = _from_dict(_to_dict(request, spider), spider)
            self._assert_same_request(request, request_copy)

            self.assertEqual(len(caught), 1)
            self.assertTrue(issubclass(caught[0].category, ScrapyDeprecationWarning))
            self.assertEqual(
                "Module scrapy.utils.reqser is deprecated, please use request.to_dict method"
                " and/or scrapy.utils.request.request_from_dict instead",
                str(caught[0].message),
            )


class TestSpiderMixin:
    def __mixin_callback(self, response):
        pass


class TestSpiderDelegation:
    def delegated_callback(self, response):
        pass


def parse_item(response):
    pass


def handle_error(failure):
    pass


def private_parse_item(response):
    pass


def private_handle_error(failure):
    pass


class TestSpider(Spider, TestSpiderMixin):
    name = 'test'
    parse_item_reference = parse_item
    handle_error_reference = handle_error
    __parse_item_reference = private_parse_item
    __handle_error_reference = private_handle_error

    def __init__(self):
        self.delegated_callback = TestSpiderDelegation().delegated_callback

    def parse_item(self, response):
        pass

    def handle_error(self, failure):
        pass

    def __parse_item_private(self, response):
        pass



import json
import os
import pstats
import shutil
import sys
import tempfile
import unittest
from io import StringIO
from subprocess import Popen, PIPE

from scrapy.utils.test import get_testenv


class CmdlineTest(unittest.TestCase):

    def setUp(self):
        self.env = get_testenv()
        self.env['SCRAPY_SETTINGS_MODULE'] = 'tests.test_cmdline.settings'

    def _execute(self, *new_args, **kwargs):
        encoding = getattr(sys.stdout, 'encoding') or 'utf-8'
        args = (sys.executable, '-m', 'scrapy.cmdline') + new_args
        proc = Popen(args, stdout=PIPE, stderr=PIPE, env=self.env, **kwargs)
        comm = proc.communicate()[0].strip()
        return comm.decode(encoding)

    def test_default_settings(self):
        self.assertEqual(self._execute('settings', '--get', 'TEST1'), 'default')

    def test_override_settings_using_set_arg(self):
        self.assertEqual(self._execute('settings', '--get', 'TEST1', '-s',
                                       'TEST1=override'), 'override')

    def test_override_settings_using_envvar(self):
        self.env['SCRAPY_TEST1'] = 'override'
        self.assertEqual(self._execute('settings', '--get', 'TEST1'), 'override')

    def test_profiling(self):
        path = tempfile.mkdtemp()
        filename = os.path.join(path, 'res.prof')
        try:
            self._execute('version', '--profile', filename)
            self.assertTrue(os.path.exists(filename))
            out = StringIO()
            stats = pstats.Stats(filename, stream=out)
            stats.print_stats()
            out.seek(0)
            stats = out.read()
            self.assertIn(os.path.join('scrapy', 'commands', 'version.py'),
                          stats)
            self.assertIn('tottime', stats)
        finally:
            shutil.rmtree(path)

    def test_override_dict_settings(self):
        EXT_PATH = "tests.test_cmdline.extensions.DummyExtension"
        EXTENSIONS = {EXT_PATH: 200}
        settingsstr = self._execute('settings', '--get', 'EXTENSIONS', '-s',
                                    'EXTENSIONS=' + json.dumps(EXTENSIONS))
        # XXX: There's gotta be a smarter way to do this...
        self.assertNotIn("...", settingsstr)
        for char in ("'", "<", ">"):
            settingsstr = settingsstr.replace(char, '"')
        settingsdict = json.loads(settingsstr)
        self.assertCountEqual(settingsdict.keys(), EXTENSIONS.keys())
        self.assertEqual(200, settingsdict[EXT_PATH])



import json
import random


def open_ua_file(path="./ua.json"):
    with open(path, 'r') as f:
        s = f.read()
    return json.loads(s)


def choice_brower_model(browser='chrome'):
    """
    browser :选择浏览器头
    默认选择:chrome
    其他选择:"opera"/"firefox"/"internetexplorer"/"safari"
    或者任意ua:"random"
    """
    try:
        ua_dic = open_ua_file()  # 弃用
    except:
        ua_dic = ua_json
    if browser.lower() in ['chrome', 'opera', 'firefox', 'internetexplorer', 'safari']:
        return random.choice(ua_dic.get("browsers").get(browser.lower()))
    else:
        return choice_random_model(ua_dic)


def choice_random_model(data):
    data_randomize = list(data['randomize'].values())
    browser = random.choice(['chrome', 'opera', 'firefox', 'internetexplorer', 'safari'])
    return choice_brower_model(browser)


ua_json = {
    "browsers": {
        "chrome": ["Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/78.0.3904.108 Safari/537.36 QIHU 360EE",
                   "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.88 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/41.0.2228.0 Safari/537.36",
                   "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/41.0.2227.1 Safari/537.36",
                   "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/41.0.2227.0 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/41.0.2227.0 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/41.0.2226.0 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 6.4; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/41.0.2225.0 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/41.0.2225.0 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 5.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/41.0.2224.3 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/40.0.2214.93 Safari/537.36",
                   "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/37.0.2062.124 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/37.0.2049.0 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 4.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/37.0.2049.0 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/36.0.1985.67 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 5.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/36.0.1985.67 Safari/537.36",
                   "Mozilla/5.0 (X11; OpenBSD i386) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/36.0.1985.125 Safari/537.36",
                   "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/36.0.1944.0 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 5.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/35.0.3319.102 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 5.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/35.0.2309.372 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 5.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/35.0.2117.157 Safari/537.36",
                   "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/35.0.1916.47 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 5.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/34.0.1866.237 Safari/537.36",
                   "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/34.0.1847.137 Safari/4E423F",
                   "Mozilla/5.0 (Windows NT 5.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/34.0.1847.116 Safari/537.36 Mozilla/5.0 (iPad; U; CPU OS 3_2 like Mac OS X; en-us) AppleWebKit/531.21.10 (KHTML, like Gecko) Version/4.0.4 Mobile/7B334b Safari/531.21.10",
                   "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/33.0.1750.517 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 6.2; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/32.0.1667.0 Safari/537.36",
                   "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/32.0.1664.3 Safari/537.36",
                   "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_8_0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/32.0.1664.3 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 5.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/31.0.1650.16 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/31.0.1623.0 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 6.2; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/30.0.1599.17 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/29.0.1547.62 Safari/537.36",
                   "Mozilla/5.0 (X11; CrOS i686 4319.74.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/29.0.1547.57 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 6.2; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/29.0.1547.2 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/28.0.1468.0 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 6.2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/28.0.1467.0 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 6.2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/28.0.1464.0 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 6.2; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/27.0.1500.55 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 6.2; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/27.0.1453.93 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/27.0.1453.93 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/27.0.1453.93 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 5.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/27.0.1453.93 Safari/537.36",
                   "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_8_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/27.0.1453.93 Safari/537.36",
                   "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_7_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/27.0.1453.93 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/27.0.1453.90 Safari/537.36",
                   "Mozilla/5.0 (X11; NetBSD) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/27.0.1453.116 Safari/537.36",
                   "Mozilla/5.0 (X11; CrOS i686 3912.101.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/27.0.1453.116 Safari/537.36",
                   "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.17 (KHTML, like Gecko) Chrome/24.0.1312.60 Safari/537.17",
                   "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_8_2) AppleWebKit/537.17 (KHTML, like Gecko) Chrome/24.0.1309.0 Safari/537.17",
                   "Mozilla/5.0 (Windows NT 6.2; WOW64) AppleWebKit/537.15 (KHTML, like Gecko) Chrome/24.0.1295.0 Safari/537.15",
                   "Mozilla/5.0 (Windows NT 6.2; WOW64) AppleWebKit/537.14 (KHTML, like Gecko) Chrome/24.0.1292.0 Safari/537.14"
                   ],
        "opera": [
            "Opera/9.80 (X11; Linux i686; Ubuntu/14.10) Presto/2.12.388 Version/12.16",
            "Opera/9.80 (Windows NT 6.0) Presto/2.12.388 Version/12.14",
            "Mozilla/5.0 (Windows NT 6.0; rv:2.0) Gecko/20100101 Firefox/4.0 Opera 12.14",
            "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.0) Opera 12.14",
            "Opera/12.80 (Windows NT 5.1; U; en) Presto/2.10.289 Version/12.02",
            "Opera/9.80 (Windows NT 6.1; U; es-ES) Presto/2.9.181 Version/12.00",
            "Opera/9.80 (Windows NT 5.1; U; zh-sg) Presto/2.9.181 Version/12.00",
            "Opera/12.0(Windows NT 5.2;U;en)Presto/22.9.168 Version/12.00",
            "Opera/12.0(Windows NT 5.1;U;en)Presto/22.9.168 Version/12.00",
            "Mozilla/5.0 (Windows NT 5.1) Gecko/20100101 Firefox/14.0 Opera/12.0",
            "Opera/9.80 (Windows NT 6.1; WOW64; U; pt) Presto/2.10.229 Version/11.62",
            "Opera/9.80 (Windows NT 6.0; U; pl) Presto/2.10.229 Version/11.62",
            "Opera/9.80 (Macintosh; Intel Mac OS X 10.6.8; U; fr) Presto/2.9.168 Version/11.52",
            "Opera/9.80 (Macintosh; Intel Mac OS X 10.6.8; U; de) Presto/2.9.168 Version/11.52",
            "Opera/9.80 (Windows NT 5.1; U; en) Presto/2.9.168 Version/11.51",
            "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; de) Opera 11.51",
            "Opera/9.80 (X11; Linux x86_64; U; fr) Presto/2.9.168 Version/11.50",
            "Opera/9.80 (X11; Linux i686; U; hu) Presto/2.9.168 Version/11.50",
            "Opera/9.80 (X11; Linux i686; U; ru) Presto/2.8.131 Version/11.11",
            "Opera/9.80 (X11; Linux i686; U; es-ES) Presto/2.8.131 Version/11.11",
            "Mozilla/5.0 (Windows NT 5.1; U; en; rv:1.8.1) Gecko/20061208 Firefox/5.0 Opera 11.11",
            "Opera/9.80 (X11; Linux x86_64; U; bg) Presto/2.8.131 Version/11.10",
            "Opera/9.80 (Windows NT 6.0; U; en) Presto/2.8.99 Version/11.10",
            "Opera/9.80 (Windows NT 5.1; U; zh-tw) Presto/2.8.131 Version/11.10",
            "Opera/9.80 (Windows NT 6.1; Opera Tablet/15165; U; en) Presto/2.8.149 Version/11.1",
            "Opera/9.80 (X11; Linux x86_64; U; Ubuntu/10.10 (maverick); pl) Presto/2.7.62 Version/11.01",
            "Opera/9.80 (X11; Linux i686; U; ja) Presto/2.7.62 Version/11.01",
            "Opera/9.80 (X11; Linux i686; U; fr) Presto/2.7.62 Version/11.01",
            "Opera/9.80 (Windows NT 6.1; U; zh-tw) Presto/2.7.62 Version/11.01",
            "Opera/9.80 (Windows NT 6.1; U; zh-cn) Presto/2.7.62 Version/11.01",
            "Opera/9.80 (Windows NT 6.1; U; sv) Presto/2.7.62 Version/11.01",
            "Opera/9.80 (Windows NT 6.1; U; en-US) Presto/2.7.62 Version/11.01",
            "Opera/9.80 (Windows NT 6.1; U; cs) Presto/2.7.62 Version/11.01",
            "Opera/9.80 (Windows NT 6.0; U; pl) Presto/2.7.62 Version/11.01",
            "Opera/9.80 (Windows NT 5.2; U; ru) Presto/2.7.62 Version/11.01",
            "Opera/9.80 (Windows NT 5.1; U;) Presto/2.7.62 Version/11.01",
            "Opera/9.80 (Windows NT 5.1; U; cs) Presto/2.7.62 Version/11.01",
            "Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US; rv:1.9.2.13) Gecko/20101213 Opera/9.80 (Windows NT 6.1; U; zh-tw) Presto/2.7.62 Version/11.01",
            "Mozilla/5.0 (Windows NT 6.1; U; nl; rv:1.9.1.6) Gecko/20091201 Firefox/3.5.6 Opera 11.01",
            "Mozilla/5.0 (Windows NT 6.1; U; de; rv:1.9.1.6) Gecko/20091201 Firefox/3.5.6 Opera 11.01",
            "Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 6.1; de) Opera 11.01",
            "Opera/9.80 (X11; Linux x86_64; U; pl) Presto/2.7.62 Version/11.00",
            "Opera/9.80 (X11; Linux i686; U; it) Presto/2.7.62 Version/11.00",
            "Opera/9.80 (Windows NT 6.1; U; zh-cn) Presto/2.6.37 Version/11.00",
            "Opera/9.80 (Windows NT 6.1; U; pl) Presto/2.7.62 Version/11.00",
            "Opera/9.80 (Windows NT 6.1; U; ko) Presto/2.7.62 Version/11.00",
            "Opera/9.80 (Windows NT 6.1; U; fi) Presto/2.7.62 Version/11.00",
            "Opera/9.80 (Windows NT 6.1; U; en-GB) Presto/2.7.62 Version/11.00",
            "Opera/9.80 (Windows NT 6.1 x64; U; en) Presto/2.7.62 Version/11.00",
            "Opera/9.80 (Windows NT 6.0; U; en) Presto/2.7.39 Version/11.00"
        ],
        "firefox": [
            "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:40.0) Gecko/20100101 Firefox/40.1",
            "Mozilla/5.0 (Windows NT 6.3; rv:36.0) Gecko/20100101 Firefox/36.0",
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10; rv:33.0) Gecko/20100101 Firefox/33.0",
            "Mozilla/5.0 (X11; Linux i586; rv:31.0) Gecko/20100101 Firefox/31.0",
            "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:31.0) Gecko/20130401 Firefox/31.0",
            "Mozilla/5.0 (Windows NT 5.1; rv:31.0) Gecko/20100101 Firefox/31.0",
            "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:29.0) Gecko/20120101 Firefox/29.0",
            "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:25.0) Gecko/20100101 Firefox/29.0",
            "Mozilla/5.0 (X11; OpenBSD amd64; rv:28.0) Gecko/20100101 Firefox/28.0",
            "Mozilla/5.0 (X11; Linux x86_64; rv:28.0) Gecko/20100101  Firefox/28.0",
            "Mozilla/5.0 (Windows NT 6.1; rv:27.3) Gecko/20130101 Firefox/27.3",
            "Mozilla/5.0 (Windows NT 6.2; Win64; x64; rv:27.0) Gecko/20121011 Firefox/27.0",
            "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:25.0) Gecko/20100101 Firefox/25.0",
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.6; rv:25.0) Gecko/20100101 Firefox/25.0",
            "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:24.0) Gecko/20100101 Firefox/24.0",
            "Mozilla/5.0 (Windows NT 6.0; WOW64; rv:24.0) Gecko/20100101 Firefox/24.0",
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.8; rv:24.0) Gecko/20100101 Firefox/24.0",
            "Mozilla/5.0 (Windows NT 6.2; rv:22.0) Gecko/20130405 Firefox/23.0",
            "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:23.0) Gecko/20130406 Firefox/23.0",
            "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:23.0) Gecko/20131011 Firefox/23.0",
            "Mozilla/5.0 (Windows NT 6.2; rv:22.0) Gecko/20130405 Firefox/22.0",
            "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:22.0) Gecko/20130328 Firefox/22.0",
            "Mozilla/5.0 (Windows NT 6.1; rv:22.0) Gecko/20130405 Firefox/22.0",
            "Mozilla/5.0 (Microsoft Windows NT 6.2.9200.0); rv:22.0) Gecko/20130405 Firefox/22.0",
            "Mozilla/5.0 (Windows NT 6.2; Win64; x64; rv:16.0.1) Gecko/20121011 Firefox/21.0.1",
            "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:16.0.1) Gecko/20121011 Firefox/21.0.1",
            "Mozilla/5.0 (Windows NT 6.2; Win64; x64; rv:21.0.0) Gecko/20121011 Firefox/21.0.0",
            "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:21.0) Gecko/20130331 Firefox/21.0",
            "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:21.0) Gecko/20100101 Firefox/21.0",
            "Mozilla/5.0 (X11; Linux i686; rv:21.0) Gecko/20100101 Firefox/21.0",
            "Mozilla/5.0 (Windows NT 6.2; WOW64; rv:21.0) Gecko/20130514 Firefox/21.0",
            "Mozilla/5.0 (Windows NT 6.2; rv:21.0) Gecko/20130326 Firefox/21.0",
            "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:21.0) Gecko/20130401 Firefox/21.0",
            "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:21.0) Gecko/20130331 Firefox/21.0",
            "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:21.0) Gecko/20130330 Firefox/21.0",
            "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:21.0) Gecko/20100101 Firefox/21.0",
            "Mozilla/5.0 (Windows NT 6.1; rv:21.0) Gecko/20130401 Firefox/21.0",
            "Mozilla/5.0 (Windows NT 6.1; rv:21.0) Gecko/20130328 Firefox/21.0",
            "Mozilla/5.0 (Windows NT 6.1; rv:21.0) Gecko/20100101 Firefox/21.0",
            "Mozilla/5.0 (Windows NT 5.1; rv:21.0) Gecko/20130401 Firefox/21.0",
            "Mozilla/5.0 (Windows NT 5.1; rv:21.0) Gecko/20130331 Firefox/21.0",
            "Mozilla/5.0 (Windows NT 5.1; rv:21.0) Gecko/20100101 Firefox/21.0",
            "Mozilla/5.0 (Windows NT 5.0; rv:21.0) Gecko/20100101 Firefox/21.0",
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.8; rv:21.0) Gecko/20100101 Firefox/21.0",
            "Mozilla/5.0 (Windows NT 6.2; Win64; x64;) Gecko/20100101 Firefox/20.0",
            "Mozilla/5.0 (Windows x86; rv:19.0) Gecko/20100101 Firefox/19.0",
            "Mozilla/5.0 (Windows NT 6.1; rv:6.0) Gecko/20100101 Firefox/19.0",
            "Mozilla/5.0 (Windows NT 6.1; rv:14.0) Gecko/20100101 Firefox/18.0.1",
            "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:18.0)  Gecko/20100101 Firefox/18.0",
            "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:17.0) Gecko/20100101 Firefox/17.0.6"
        ],
        "internetexplorer": [
            "Mozilla/5.0 (Windows NT 6.1; WOW64; Trident/7.0; AS; rv:11.0) like Gecko",
            "Mozilla/5.0 (compatible, MSIE 11, Windows NT 6.3; Trident/7.0;  rv:11.0) like Gecko",
            "Mozilla/5.0 (compatible; MSIE 10.6; Windows NT 6.1; Trident/5.0; InfoPath.2; SLCC1; .NET CLR 3.0.4506.2152; .NET CLR 3.5.30729; .NET CLR 2.0.50727) 3gpp-gba UNTRUSTED/1.0",
            "Mozilla/5.0 (compatible; MSIE 10.0; Windows NT 7.0; InfoPath.3; .NET CLR 3.1.40767; Trident/6.0; en-IN)",
            "Mozilla/5.0 (compatible; MSIE 10.0; Windows NT 6.1; WOW64; Trident/6.0)",
            "Mozilla/5.0 (compatible; MSIE 10.0; Windows NT 6.1; Trident/6.0)",
            "Mozilla/5.0 (compatible; MSIE 10.0; Windows NT 6.1; Trident/5.0)",
            "Mozilla/5.0 (compatible; MSIE 10.0; Windows NT 6.1; Trident/4.0; InfoPath.2; SV1; .NET CLR 2.0.50727; WOW64)",
            "Mozilla/5.0 (compatible; MSIE 10.0; Macintosh; Intel Mac OS X 10_7_3; Trident/6.0)",
            "Mozilla/4.0 (Compatible; MSIE 8.0; Windows NT 5.2; Trident/6.0)",
            "Mozilla/4.0 (compatible; MSIE 10.0; Windows NT 6.1; Trident/5.0)",
            "Mozilla/1.22 (compatible; MSIE 10.0; Windows 3.1)",
            "Mozilla/5.0 (Windows; U; MSIE 9.0; WIndows NT 9.0; en-US))",
            "Mozilla/5.0 (Windows; U; MSIE 9.0; Windows NT 9.0; en-US)",
            "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 7.1; Trident/5.0)",
            "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; WOW64; Trident/5.0; SLCC2; Media Center PC 6.0; InfoPath.3; MS-RTC LM 8; Zune 4.7)",
            "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; WOW64; Trident/5.0; SLCC2; Media Center PC 6.0; InfoPath.3; MS-RTC LM 8; Zune 4.7",
            "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; WOW64; Trident/5.0; SLCC2; .NET CLR 2.0.50727; .NET CLR 3.5.30729; .NET CLR 3.0.30729; Media Center PC 6.0; Zune 4.0; InfoPath.3; MS-RTC LM 8; .NET4.0C; .NET4.0E)",
            "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; WOW64; Trident/5.0; chromeframe/12.0.742.112)",
            "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; WOW64; Trident/5.0; .NET CLR 3.5.30729; .NET CLR 3.0.30729; .NET CLR 2.0.50727; Media Center PC 6.0)",
            "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; Win64; x64; Trident/5.0; .NET CLR 3.5.30729; .NET CLR 3.0.30729; .NET CLR 2.0.50727; Media Center PC 6.0)",
            "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; Win64; x64; Trident/5.0; .NET CLR 2.0.50727; SLCC2; .NET CLR 3.5.30729; .NET CLR 3.0.30729; Media Center PC 6.0; Zune 4.0; Tablet PC 2.0; InfoPath.3; .NET4.0C; .NET4.0E)",
            "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; Win64; x64; Trident/5.0",
            "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; Trident/5.0; yie8)",
            "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; Trident/5.0; SLCC2; .NET CLR 2.0.50727; .NET CLR 3.5.30729; .NET CLR 3.0.30729; Media Center PC 6.0; InfoPath.2; .NET CLR 1.1.4322; .NET4.0C; Tablet PC 2.0)",
            "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; Trident/5.0; FunWebProducts)",
            "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; Trident/5.0; chromeframe/13.0.782.215)",
            "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; Trident/5.0; chromeframe/11.0.696.57)",
            "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; Trident/5.0) chromeframe/10.0.648.205",
            "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; Trident/4.0; GTB7.4; InfoPath.1; SV1; .NET CLR 2.8.52393; WOW64; en-US)",
            "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.0; Trident/5.0; chromeframe/11.0.696.57)",
            "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.0; Trident/4.0; GTB7.4; InfoPath.3; SV1; .NET CLR 3.1.76908; WOW64; en-US)",
            "Mozilla/5.0 (compatible; MSIE 8.0; Windows NT 6.1; Trident/4.0; GTB7.4; InfoPath.2; SV1; .NET CLR 3.3.69573; WOW64; en-US)",
            "Mozilla/5.0 (compatible; MSIE 8.0; Windows NT 6.0; Trident/4.0; WOW64; Trident/4.0; SLCC2; .NET CLR 2.0.50727; .NET CLR 3.5.30729; .NET CLR 3.0.30729; .NET CLR 1.0.3705; .NET CLR 1.1.4322)",
            "Mozilla/5.0 (compatible; MSIE 8.0; Windows NT 6.0; Trident/4.0; InfoPath.1; SV1; .NET CLR 3.8.36217; WOW64; en-US)",
            "Mozilla/5.0 (compatible; MSIE 8.0; Windows NT 6.0; Trident/4.0; .NET CLR 2.7.58687; SLCC2; Media Center PC 5.0; Zune 3.4; Tablet PC 3.6; InfoPath.3)",
            "Mozilla/5.0 (compatible; MSIE 8.0; Windows NT 5.2; Trident/4.0; Media Center PC 4.0; SLCC1; .NET CLR 3.0.04320)",
            "Mozilla/5.0 (compatible; MSIE 8.0; Windows NT 5.1; Trident/4.0; SLCC1; .NET CLR 3.0.4506.2152; .NET CLR 3.5.30729; .NET CLR 1.1.4322)",
            "Mozilla/5.0 (compatible; MSIE 8.0; Windows NT 5.1; Trident/4.0; InfoPath.2; SLCC1; .NET CLR 3.0.4506.2152; .NET CLR 3.5.30729; .NET CLR 2.0.50727)",
            "Mozilla/5.0 (compatible; MSIE 8.0; Windows NT 5.1; Trident/4.0; .NET CLR 1.1.4322; .NET CLR 2.0.50727)",
            "Mozilla/5.0 (compatible; MSIE 8.0; Windows NT 5.1; SLCC1; .NET CLR 1.1.4322)",
            "Mozilla/5.0 (compatible; MSIE 8.0; Windows NT 5.0; Trident/4.0; InfoPath.1; SV1; .NET CLR 3.0.4506.2152; .NET CLR 3.5.30729; .NET CLR 3.0.04506.30)",
            "Mozilla/5.0 (compatible; MSIE 7.0; Windows NT 5.0; Trident/4.0; FBSMTWB; .NET CLR 2.0.34861; .NET CLR 3.0.3746.3218; .NET CLR 3.5.33652; msn OptimizedIE8;ENUS)",
            "Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 6.2; Trident/4.0; SLCC2; .NET CLR 2.0.50727; .NET CLR 3.5.30729; .NET CLR 3.0.30729; Media Center PC 6.0)",
            "Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 6.1; WOW64; Trident/4.0; SLCC2; Media Center PC 6.0; InfoPath.2; MS-RTC LM 8)",
            "Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 6.1; WOW64; Trident/4.0; SLCC2; Media Center PC 6.0; InfoPath.2; MS-RTC LM 8",
            "Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 6.1; WOW64; Trident/4.0; SLCC2; .NET CLR 2.0.50727; Media Center PC 6.0; .NET CLR 3.5.30729; .NET CLR 3.0.30729; .NET4.0C)",
            "Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 6.1; WOW64; Trident/4.0; SLCC2; .NET CLR 2.0.50727; InfoPath.3; .NET4.0C; .NET4.0E; .NET CLR 3.5.30729; .NET CLR 3.0.30729; MS-RTC LM 8)",
            "Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 6.1; WOW64; Trident/4.0; SLCC2; .NET CLR 2.0.50727; InfoPath.2)",
            "Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 6.1; WOW64; Trident/4.0; SLCC2; .NET CLR 2.0.50727; .NET CLR 3.5.30729; .NET CLR 3.0.30729; Media Center PC 6.0; Zune 3.0)"
        ],
        "safari": [
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_3) AppleWebKit/537.75.14 (KHTML, like Gecko) Version/7.0.3 Safari/7046A194A",
            "Mozilla/5.0 (iPad; CPU OS 6_0 like Mac OS X) AppleWebKit/536.26 (KHTML, like Gecko) Version/6.0 Mobile/10A5355d Safari/8536.25",
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_6_8) AppleWebKit/537.13+ (KHTML, like Gecko) Version/5.1.7 Safari/534.57.2",
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_7_3) AppleWebKit/534.55.3 (KHTML, like Gecko) Version/5.1.3 Safari/534.53.10",
            "Mozilla/5.0 (iPad; CPU OS 5_1 like Mac OS X) AppleWebKit/534.46 (KHTML, like Gecko ) Version/5.1 Mobile/9B176 Safari/7534.48.3",
            "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_6_8; de-at) AppleWebKit/533.21.1 (KHTML, like Gecko) Version/5.0.5 Safari/533.21.1",
            "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_6_7; da-dk) AppleWebKit/533.21.1 (KHTML, like Gecko) Version/5.0.5 Safari/533.21.1",
            "Mozilla/5.0 (Windows; U; Windows NT 6.1; tr-TR) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.4 Safari/533.20.27",
            "Mozilla/5.0 (Windows; U; Windows NT 6.1; ko-KR) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.4 Safari/533.20.27",
            "Mozilla/5.0 (Windows; U; Windows NT 6.1; fr-FR) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.4 Safari/533.20.27",
            "Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.4 Safari/533.20.27",
            "Mozilla/5.0 (Windows; U; Windows NT 6.1; cs-CZ) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.4 Safari/533.20.27",
            "Mozilla/5.0 (Windows; U; Windows NT 6.0; ja-JP) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.4 Safari/533.20.27",
            "Mozilla/5.0 (Windows; U; Windows NT 6.0; en-US) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.4 Safari/533.20.27",
            "Mozilla/5.0 (Macintosh; U; PPC Mac OS X 10_5_8; zh-cn) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.4 Safari/533.20.27",
            "Mozilla/5.0 (Macintosh; U; PPC Mac OS X 10_5_8; ja-jp) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.4 Safari/533.20.27",
            "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_6_7; ja-jp) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.4 Safari/533.20.27",
            "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_6_6; zh-cn) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.4 Safari/533.20.27",
            "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_6_6; sv-se) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.4 Safari/533.20.27",
            "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_6_6; ko-kr) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.4 Safari/533.20.27",
            "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_6_6; ja-jp) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.4 Safari/533.20.27",
            "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_6_6; it-it) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.4 Safari/533.20.27",
            "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_6_6; fr-fr) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.4 Safari/533.20.27",
            "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_6_6; es-es) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.4 Safari/533.20.27",
            "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_6_6; en-us) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.4 Safari/533.20.27",
            "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_6_6; en-gb) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.4 Safari/533.20.27",
            "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_6_6; de-de) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.4 Safari/533.20.27",
            "Mozilla/5.0 (Windows; U; Windows NT 6.1; sv-SE) AppleWebKit/533.19.4 (KHTML, like Gecko) Version/5.0.3 Safari/533.19.4",
            "Mozilla/5.0 (Windows; U; Windows NT 6.1; ja-JP) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.3 Safari/533.19.4",
            "Mozilla/5.0 (Windows; U; Windows NT 6.1; de-DE) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.3 Safari/533.19.4",
            "Mozilla/5.0 (Windows; U; Windows NT 6.0; hu-HU) AppleWebKit/533.19.4 (KHTML, like Gecko) Version/5.0.3 Safari/533.19.4",
            "Mozilla/5.0 (Windows; U; Windows NT 6.0; en-US) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.3 Safari/533.19.4",
            "Mozilla/5.0 (Windows; U; Windows NT 6.0; de-DE) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.3 Safari/533.19.4",
            "Mozilla/5.0 (Windows; U; Windows NT 5.1; ru-RU) AppleWebKit/533.19.4 (KHTML, like Gecko) Version/5.0.3 Safari/533.19.4",
            "Mozilla/5.0 (Windows; U; Windows NT 5.1; ja-JP) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.3 Safari/533.19.4",
            "Mozilla/5.0 (Windows; U; Windows NT 5.1; it-IT) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.3 Safari/533.19.4",
            "Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US) AppleWebKit/533.20.25 (KHTML, like Gecko) Version/5.0.3 Safari/533.19.4",
            "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_6_7; en-us) AppleWebKit/534.16+ (KHTML, like Gecko) Version/5.0.3 Safari/533.19.4",
            "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_6_6; fr-ch) AppleWebKit/533.19.4 (KHTML, like Gecko) Version/5.0.3 Safari/533.19.4",
            "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_6_5; de-de) AppleWebKit/534.15+ (KHTML, like Gecko) Version/5.0.3 Safari/533.19.4",
            "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_6_5; ar) AppleWebKit/533.19.4 (KHTML, like Gecko) Version/5.0.3 Safari/533.19.4",
            "Mozilla/5.0 (Android 2.2; Windows; U; Windows NT 6.1; en-US) AppleWebKit/533.19.4 (KHTML, like Gecko) Version/5.0.3 Safari/533.19.4",
            "Mozilla/5.0 (Windows; U; Windows NT 6.1; zh-HK) AppleWebKit/533.18.1 (KHTML, like Gecko) Version/5.0.2 Safari/533.18.5",
            "Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US) AppleWebKit/533.19.4 (KHTML, like Gecko) Version/5.0.2 Safari/533.18.5",
            "Mozilla/5.0 (Windows; U; Windows NT 6.0; tr-TR) AppleWebKit/533.18.1 (KHTML, like Gecko) Version/5.0.2 Safari/533.18.5",
            "Mozilla/5.0 (Windows; U; Windows NT 6.0; nb-NO) AppleWebKit/533.18.1 (KHTML, like Gecko) Version/5.0.2 Safari/533.18.5",
            "Mozilla/5.0 (Windows; U; Windows NT 6.0; fr-FR) AppleWebKit/533.18.1 (KHTML, like Gecko) Version/5.0.2 Safari/533.18.5",
            "Mozilla/5.0 (Windows; U; Windows NT 5.1; zh-TW) AppleWebKit/533.19.4 (KHTML, like Gecko) Version/5.0.2 Safari/533.18.5",
            "Mozilla/5.0 (Windows; U; Windows NT 5.1; ru-RU) AppleWebKit/533.18.1 (KHTML, like Gecko) Version/5.0.2 Safari/533.18.5",
            "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_5_8; zh-cn) AppleWebKit/533.18.1 (KHTML, like Gecko) Version/5.0.2 Safari/533.18.5"
        ]
    },
    "randomize": {
        "344": "chrome",
        "819": "firefox",
        "346": "chrome",
        "347": "chrome",
        "340": "chrome",
        "341": "chrome",
        "342": "chrome",
        "343": "chrome",
        "810": "internetexplorer",
        "811": "internetexplorer",
        "812": "internetexplorer",
        "813": "firefox",
        "348": "chrome",
        "349": "chrome",
        "816": "firefox",
        "817": "firefox",
        "737": "chrome",
        "719": "chrome",
        "718": "chrome",
        "717": "chrome",
        "716": "chrome",
        "715": "chrome",
        "714": "chrome",
        "713": "chrome",
        "712": "chrome",
        "711": "chrome",
        "710": "chrome",
        "421": "chrome",
        "129": "chrome",
        "420": "chrome",
        "423": "chrome",
        "422": "chrome",
        "425": "chrome",
        "619": "chrome",
        "424": "chrome",
        "427": "chrome",
        "298": "chrome",
        "299": "chrome",
        "296": "chrome",
        "297": "chrome",
        "294": "chrome",
        "295": "chrome",
        "292": "chrome",
        "293": "chrome",
        "290": "chrome",
        "291": "chrome",
        "591": "chrome",
        "590": "chrome",
        "593": "chrome",
        "592": "chrome",
        "595": "chrome",
        "594": "chrome",
        "597": "chrome",
        "596": "chrome",
        "195": "chrome",
        "194": "chrome",
        "197": "chrome",
        "196": "chrome",
        "191": "chrome",
        "190": "chrome",
        "193": "chrome",
        "192": "chrome",
        "270": "chrome",
        "271": "chrome",
        "272": "chrome",
        "273": "chrome",
        "274": "chrome",
        "275": "chrome",
        "276": "chrome",
        "277": "chrome",
        "278": "chrome",
        "279": "chrome",
        "569": "chrome",
        "497": "chrome",
        "524": "chrome",
        "525": "chrome",
        "526": "chrome",
        "527": "chrome",
        "520": "chrome",
        "521": "chrome",
        "522": "chrome",
        "523": "chrome",
        "528": "chrome",
        "529": "chrome",
        "449": "chrome",
        "448": "chrome",
        "345": "chrome",
        "443": "chrome",
        "442": "chrome",
        "441": "chrome",
        "440": "chrome",
        "447": "chrome",
        "446": "chrome",
        "445": "chrome",
        "444": "chrome",
        "47": "chrome",
        "108": "chrome",
        "109": "chrome",
        "102": "chrome",
        "103": "chrome",
        "100": "chrome",
        "101": "chrome",
        "106": "chrome",
        "107": "chrome",
        "104": "chrome",
        "105": "chrome",
        "902": "firefox",
        "903": "firefox",
        "39": "chrome",
        "38": "chrome",
        "906": "firefox",
        "907": "firefox",
        "904": "firefox",
        "905": "firefox",
        "33": "chrome",
        "32": "chrome",
        "31": "chrome",
        "30": "chrome",
        "37": "chrome",
        "36": "chrome",
        "35": "chrome",
        "34": "chrome",
        "641": "chrome",
        "640": "chrome",
        "643": "chrome",
        "642": "chrome",
        "645": "chrome",
        "644": "chrome",
        "438": "chrome",
        "439": "chrome",
        "436": "chrome",
        "437": "chrome",
        "434": "chrome",
        "435": "chrome",
        "432": "chrome",
        "433": "chrome",
        "430": "chrome",
        "431": "chrome",
        "826": "firefox",
        "339": "chrome",
        "338": "chrome",
        "335": "chrome",
        "334": "chrome",
        "337": "chrome",
        "336": "chrome",
        "331": "chrome",
        "330": "chrome",
        "333": "chrome",
        "332": "chrome",
        "559": "chrome",
        "745": "chrome",
        "854": "firefox",
        "818": "firefox",
        "856": "firefox",
        "857": "firefox",
        "850": "firefox",
        "851": "firefox",
        "852": "firefox",
        "0": "chrome",
        "858": "firefox",
        "859": "firefox",
        "748": "chrome",
        "6": "chrome",
        "43": "chrome",
        "99": "chrome",
        "98": "chrome",
        "91": "chrome",
        "90": "chrome",
        "93": "chrome",
        "92": "chrome",
        "95": "chrome",
        "94": "chrome",
        "97": "chrome",
        "96": "chrome",
        "814": "firefox",
        "815": "firefox",
        "153": "chrome",
        "740": "chrome",
        "741": "chrome",
        "742": "chrome",
        "743": "chrome",
        "744": "chrome",
        "558": "chrome",
        "746": "chrome",
        "747": "chrome",
        "555": "chrome",
        "554": "chrome",
        "557": "chrome",
        "556": "chrome",
        "551": "chrome",
        "550": "chrome",
        "553": "chrome",
        "552": "chrome",
        "238": "chrome",
        "239": "chrome",
        "234": "chrome",
        "235": "chrome",
        "236": "chrome",
        "237": "chrome",
        "230": "chrome",
        "231": "chrome",
        "232": "chrome",
        "233": "chrome",
        "1": "chrome",
        "155": "chrome",
        "146": "chrome",
        "147": "chrome",
        "618": "chrome",
        "145": "chrome",
        "142": "chrome",
        "143": "chrome",
        "140": "chrome",
        "141": "chrome",
        "612": "chrome",
        "613": "chrome",
        "610": "chrome",
        "611": "chrome",
        "616": "chrome",
        "617": "chrome",
        "148": "chrome",
        "149": "chrome",
        "46": "chrome",
        "154": "chrome",
        "948": "safari",
        "949": "safari",
        "946": "safari",
        "947": "safari",
        "944": "safari",
        "945": "safari",
        "942": "safari",
        "943": "safari",
        "940": "safari",
        "941": "safari",
        "689": "chrome",
        "688": "chrome",
        "685": "chrome",
        "684": "chrome",
        "687": "chrome",
        "686": "chrome",
        "681": "chrome",
        "680": "chrome",
        "683": "chrome",
        "682": "chrome",
        "458": "chrome",
        "459": "chrome",
        "133": "chrome",
        "132": "chrome",
        "131": "chrome",
        "130": "chrome",
        "137": "chrome",
        "136": "chrome",
        "135": "chrome",
        "134": "chrome",
        "494": "chrome",
        "495": "chrome",
        "139": "chrome",
        "138": "chrome",
        "490": "chrome",
        "491": "chrome",
        "492": "chrome",
        "493": "chrome",
        "24": "chrome",
        "25": "chrome",
        "26": "chrome",
        "27": "chrome",
        "20": "chrome",
        "21": "chrome",
        "22": "chrome",
        "23": "chrome",
        "28": "chrome",
        "29": "chrome",
        "820": "firefox",
        "407": "chrome",
        "406": "chrome",
        "405": "chrome",
        "404": "chrome",
        "403": "chrome",
        "402": "chrome",
        "401": "chrome",
        "400": "chrome",
        "933": "firefox",
        "932": "firefox",
        "931": "firefox",
        "930": "firefox",
        "937": "safari",
        "452": "chrome",
        "409": "chrome",
        "408": "chrome",
        "453": "chrome",
        "414": "chrome",
        "183": "chrome",
        "415": "chrome",
        "379": "chrome",
        "378": "chrome",
        "228": "chrome",
        "829": "firefox",
        "828": "firefox",
        "371": "chrome",
        "370": "chrome",
        "373": "chrome",
        "372": "chrome",
        "375": "chrome",
        "374": "chrome",
        "377": "chrome",
        "376": "chrome",
        "708": "chrome",
        "709": "chrome",
        "704": "chrome",
        "705": "chrome",
        "706": "chrome",
        "707": "chrome",
        "700": "chrome",
        "144": "chrome",
        "702": "chrome",
        "703": "chrome",
        "393": "chrome",
        "392": "chrome",
        "88": "chrome",
        "89": "chrome",
        "397": "chrome",
        "396": "chrome",
        "395": "chrome",
        "394": "chrome",
        "82": "chrome",
        "83": "chrome",
        "80": "chrome",
        "81": "chrome",
        "86": "chrome",
        "87": "chrome",
        "84": "chrome",
        "85": "chrome",
        "797": "internetexplorer",
        "796": "internetexplorer",
        "795": "internetexplorer",
        "794": "internetexplorer",
        "793": "internetexplorer",
        "792": "internetexplorer",
        "791": "internetexplorer",
        "790": "internetexplorer",
        "749": "chrome",
        "799": "internetexplorer",
        "798": "internetexplorer",
        "7": "chrome",
        "170": "chrome",
        "586": "chrome",
        "587": "chrome",
        "584": "chrome",
        "585": "chrome",
        "582": "chrome",
        "583": "chrome",
        "580": "chrome",
        "581": "chrome",
        "588": "chrome",
        "589": "chrome",
        "245": "chrome",
        "244": "chrome",
        "247": "chrome",
        "246": "chrome",
        "241": "chrome",
        "614": "chrome",
        "243": "chrome",
        "242": "chrome",
        "615": "chrome",
        "249": "chrome",
        "248": "chrome",
        "418": "chrome",
        "419": "chrome",
        "519": "chrome",
        "518": "chrome",
        "511": "chrome",
        "510": "chrome",
        "513": "chrome",
        "512": "chrome",
        "515": "chrome",
        "514": "chrome",
        "517": "chrome",
        "516": "chrome",
        "623": "chrome",
        "622": "chrome",
        "621": "chrome",
        "620": "chrome",
        "627": "chrome",
        "626": "chrome",
        "625": "chrome",
        "624": "chrome",
        "450": "chrome",
        "451": "chrome",
        "629": "chrome",
        "628": "chrome",
        "454": "chrome",
        "455": "chrome",
        "456": "chrome",
        "457": "chrome",
        "179": "chrome",
        "178": "chrome",
        "177": "chrome",
        "199": "chrome",
        "175": "chrome",
        "174": "chrome",
        "173": "chrome",
        "172": "chrome",
        "171": "chrome",
        "198": "chrome",
        "977": "opera",
        "182": "chrome",
        "975": "opera",
        "974": "opera",
        "973": "opera",
        "972": "opera",
        "971": "opera",
        "970": "opera",
        "180": "chrome",
        "979": "opera",
        "978": "opera",
        "656": "chrome",
        "599": "chrome",
        "654": "chrome",
        "181": "chrome",
        "186": "chrome",
        "187": "chrome",
        "184": "chrome",
        "185": "chrome",
        "652": "chrome",
        "188": "chrome",
        "189": "chrome",
        "658": "chrome",
        "653": "chrome",
        "650": "chrome",
        "651": "chrome",
        "11": "chrome",
        "10": "chrome",
        "13": "chrome",
        "12": "chrome",
        "15": "chrome",
        "14": "chrome",
        "17": "chrome",
        "16": "chrome",
        "19": "chrome",
        "18": "chrome",
        "863": "firefox",
        "862": "firefox",
        "865": "firefox",
        "864": "firefox",
        "867": "firefox",
        "866": "firefox",
        "354": "chrome",
        "659": "chrome",
        "44": "chrome",
        "883": "firefox",
        "882": "firefox",
        "881": "firefox",
        "880": "firefox",
        "887": "firefox",
        "886": "firefox",
        "885": "firefox",
        "884": "firefox",
        "889": "firefox",
        "888": "firefox",
        "116": "chrome",
        "45": "chrome",
        "657": "chrome",
        "355": "chrome",
        "322": "chrome",
        "323": "chrome",
        "320": "chrome",
        "321": "chrome",
        "326": "chrome",
        "327": "chrome",
        "324": "chrome",
        "325": "chrome",
        "328": "chrome",
        "329": "chrome",
        "562": "chrome",
        "201": "chrome",
        "200": "chrome",
        "203": "chrome",
        "202": "chrome",
        "205": "chrome",
        "204": "chrome",
        "207": "chrome",
        "206": "chrome",
        "209": "chrome",
        "208": "chrome",
        "779": "internetexplorer",
        "778": "internetexplorer",
        "77": "chrome",
        "76": "chrome",
        "75": "chrome",
        "74": "chrome",
        "73": "chrome",
        "72": "chrome",
        "71": "chrome",
        "70": "chrome",
        "655": "chrome",
        "567": "chrome",
        "79": "chrome",
        "78": "chrome",
        "359": "chrome",
        "358": "chrome",
        "669": "chrome",
        "668": "chrome",
        "667": "chrome",
        "666": "chrome",
        "665": "chrome",
        "664": "chrome",
        "663": "chrome",
        "662": "chrome",
        "661": "chrome",
        "660": "chrome",
        "215": "chrome",
        "692": "chrome",
        "693": "chrome",
        "690": "chrome",
        "691": "chrome",
        "696": "chrome",
        "697": "chrome",
        "694": "chrome",
        "695": "chrome",
        "698": "chrome",
        "699": "chrome",
        "542": "chrome",
        "543": "chrome",
        "540": "chrome",
        "541": "chrome",
        "546": "chrome",
        "547": "chrome",
        "544": "chrome",
        "545": "chrome",
        "8": "chrome",
        "548": "chrome",
        "549": "chrome",
        "598": "chrome",
        "869": "firefox",
        "868": "firefox",
        "120": "chrome",
        "121": "chrome",
        "122": "chrome",
        "123": "chrome",
        "124": "chrome",
        "125": "chrome",
        "126": "chrome",
        "127": "chrome",
        "128": "chrome",
        "2": "chrome",
        "219": "chrome",
        "176": "chrome",
        "214": "chrome",
        "563": "chrome",
        "928": "firefox",
        "929": "firefox",
        "416": "chrome",
        "417": "chrome",
        "410": "chrome",
        "411": "chrome",
        "412": "chrome",
        "413": "chrome",
        "920": "firefox",
        "498": "chrome",
        "922": "firefox",
        "923": "firefox",
        "924": "firefox",
        "925": "firefox",
        "926": "firefox",
        "927": "firefox",
        "319": "chrome",
        "318": "chrome",
        "313": "chrome",
        "312": "chrome",
        "311": "chrome",
        "310": "chrome",
        "317": "chrome",
        "316": "chrome",
        "315": "chrome",
        "314": "chrome",
        "921": "firefox",
        "496": "chrome",
        "832": "firefox",
        "833": "firefox",
        "830": "firefox",
        "831": "firefox",
        "836": "firefox",
        "837": "firefox",
        "834": "firefox",
        "835": "firefox",
        "838": "firefox",
        "839": "firefox",
        "3": "chrome",
        "368": "chrome",
        "369": "chrome",
        "366": "chrome",
        "367": "chrome",
        "364": "chrome",
        "365": "chrome",
        "362": "chrome",
        "363": "chrome",
        "360": "chrome",
        "361": "chrome",
        "218": "chrome",
        "380": "chrome",
        "861": "firefox",
        "382": "chrome",
        "383": "chrome",
        "384": "chrome",
        "385": "chrome",
        "386": "chrome",
        "387": "chrome",
        "388": "chrome",
        "389": "chrome",
        "784": "internetexplorer",
        "785": "internetexplorer",
        "786": "internetexplorer",
        "787": "internetexplorer",
        "780": "internetexplorer",
        "781": "internetexplorer",
        "782": "internetexplorer",
        "381": "chrome",
        "788": "internetexplorer",
        "789": "internetexplorer",
        "860": "firefox",
        "151": "chrome",
        "579": "chrome",
        "578": "chrome",
        "150": "chrome",
        "573": "chrome",
        "572": "chrome",
        "571": "chrome",
        "570": "chrome",
        "577": "chrome",
        "576": "chrome",
        "575": "chrome",
        "574": "chrome",
        "60": "chrome",
        "61": "chrome",
        "62": "chrome",
        "259": "chrome",
        "64": "chrome",
        "65": "chrome",
        "66": "chrome",
        "67": "chrome",
        "68": "chrome",
        "253": "chrome",
        "250": "chrome",
        "251": "chrome",
        "256": "chrome",
        "257": "chrome",
        "254": "chrome",
        "255": "chrome",
        "499": "chrome",
        "157": "chrome",
        "156": "chrome",
        "939": "safari",
        "731": "chrome",
        "730": "chrome",
        "733": "chrome",
        "938": "safari",
        "735": "chrome",
        "734": "chrome",
        "508": "chrome",
        "736": "chrome",
        "506": "chrome",
        "738": "chrome",
        "504": "chrome",
        "505": "chrome",
        "502": "chrome",
        "503": "chrome",
        "500": "chrome",
        "501": "chrome",
        "630": "chrome",
        "631": "chrome",
        "632": "chrome",
        "633": "chrome",
        "469": "chrome",
        "468": "chrome",
        "636": "chrome",
        "637": "chrome",
        "465": "chrome",
        "464": "chrome",
        "467": "chrome",
        "466": "chrome",
        "461": "chrome",
        "900": "firefox",
        "463": "chrome",
        "462": "chrome",
        "901": "firefox",
        "168": "chrome",
        "169": "chrome",
        "164": "chrome",
        "165": "chrome",
        "166": "chrome",
        "167": "chrome",
        "160": "chrome",
        "161": "chrome",
        "162": "chrome",
        "163": "chrome",
        "964": "safari",
        "965": "safari",
        "966": "safari",
        "967": "safari",
        "960": "safari",
        "961": "safari",
        "962": "safari",
        "963": "safari",
        "783": "internetexplorer",
        "968": "safari",
        "969": "opera",
        "936": "firefox",
        "935": "firefox",
        "934": "firefox",
        "908": "firefox",
        "909": "firefox",
        "722": "chrome",
        "426": "chrome",
        "878": "firefox",
        "879": "firefox",
        "876": "firefox",
        "877": "firefox",
        "874": "firefox",
        "875": "firefox",
        "872": "firefox",
        "873": "firefox",
        "870": "firefox",
        "871": "firefox",
        "9": "chrome",
        "890": "firefox",
        "891": "firefox",
        "892": "firefox",
        "893": "firefox",
        "894": "firefox",
        "647": "chrome",
        "896": "firefox",
        "897": "firefox",
        "898": "firefox",
        "899": "firefox",
        "646": "chrome",
        "649": "chrome",
        "648": "chrome",
        "357": "chrome",
        "356": "chrome",
        "809": "internetexplorer",
        "808": "internetexplorer",
        "353": "chrome",
        "352": "chrome",
        "351": "chrome",
        "350": "chrome",
        "803": "internetexplorer",
        "802": "internetexplorer",
        "801": "internetexplorer",
        "800": "internetexplorer",
        "807": "internetexplorer",
        "806": "internetexplorer",
        "805": "internetexplorer",
        "804": "internetexplorer",
        "216": "chrome",
        "217": "chrome",
        "768": "chrome",
        "769": "chrome",
        "212": "chrome",
        "213": "chrome",
        "210": "chrome",
        "211": "chrome",
        "762": "chrome",
        "763": "chrome",
        "760": "chrome",
        "761": "chrome",
        "766": "chrome",
        "767": "chrome",
        "764": "chrome",
        "765": "chrome",
        "40": "chrome",
        "41": "chrome",
        "289": "chrome",
        "288": "chrome",
        "4": "chrome",
        "281": "chrome",
        "280": "chrome",
        "283": "chrome",
        "282": "chrome",
        "285": "chrome",
        "284": "chrome",
        "287": "chrome",
        "286": "chrome",
        "678": "chrome",
        "679": "chrome",
        "674": "chrome",
        "675": "chrome",
        "676": "chrome",
        "677": "chrome",
        "670": "chrome",
        "671": "chrome",
        "672": "chrome",
        "673": "chrome",
        "263": "chrome",
        "262": "chrome",
        "261": "chrome",
        "260": "chrome",
        "267": "chrome",
        "266": "chrome",
        "265": "chrome",
        "264": "chrome",
        "269": "chrome",
        "268": "chrome",
        "59": "chrome",
        "58": "chrome",
        "55": "chrome",
        "54": "chrome",
        "57": "chrome",
        "56": "chrome",
        "51": "chrome",
        "258": "chrome",
        "53": "chrome",
        "52": "chrome",
        "537": "chrome",
        "536": "chrome",
        "535": "chrome",
        "63": "chrome",
        "533": "chrome",
        "532": "chrome",
        "531": "chrome",
        "530": "chrome",
        "152": "chrome",
        "539": "chrome",
        "538": "chrome",
        "775": "internetexplorer",
        "774": "internetexplorer",
        "982": "opera",
        "983": "opera",
        "980": "opera",
        "981": "opera",
        "777": "internetexplorer",
        "984": "opera",
        "50": "chrome",
        "115": "chrome",
        "252": "chrome",
        "117": "chrome",
        "776": "internetexplorer",
        "111": "chrome",
        "110": "chrome",
        "113": "chrome",
        "69": "chrome",
        "771": "chrome",
        "119": "chrome",
        "118": "chrome",
        "770": "chrome",
        "773": "internetexplorer",
        "772": "internetexplorer",
        "429": "chrome",
        "428": "chrome",
        "534": "chrome",
        "919": "firefox",
        "918": "firefox",
        "915": "firefox",
        "914": "firefox",
        "917": "firefox",
        "916": "firefox",
        "911": "firefox",
        "910": "firefox",
        "913": "firefox",
        "912": "firefox",
        "308": "chrome",
        "309": "chrome",
        "855": "firefox",
        "300": "chrome",
        "301": "chrome",
        "302": "chrome",
        "303": "chrome",
        "304": "chrome",
        "305": "chrome",
        "306": "chrome",
        "307": "chrome",
        "895": "firefox",
        "825": "firefox",
        "824": "firefox",
        "827": "firefox",
        "847": "firefox",
        "846": "firefox",
        "845": "firefox",
        "844": "firefox",
        "843": "firefox",
        "842": "firefox",
        "841": "firefox",
        "840": "firefox",
        "821": "firefox",
        "853": "firefox",
        "849": "firefox",
        "848": "firefox",
        "823": "firefox",
        "822": "firefox",
        "240": "chrome",
        "390": "chrome",
        "732": "chrome",
        "753": "chrome",
        "752": "chrome",
        "751": "chrome",
        "750": "chrome",
        "757": "chrome",
        "756": "chrome",
        "755": "chrome",
        "754": "chrome",
        "560": "chrome",
        "561": "chrome",
        "759": "chrome",
        "758": "chrome",
        "564": "chrome",
        "565": "chrome",
        "566": "chrome",
        "701": "chrome",
        "739": "chrome",
        "229": "chrome",
        "507": "chrome",
        "227": "chrome",
        "226": "chrome",
        "225": "chrome",
        "224": "chrome",
        "223": "chrome",
        "222": "chrome",
        "221": "chrome",
        "220": "chrome",
        "114": "chrome",
        "391": "chrome",
        "726": "chrome",
        "727": "chrome",
        "724": "chrome",
        "725": "chrome",
        "568": "chrome",
        "723": "chrome",
        "720": "chrome",
        "721": "chrome",
        "728": "chrome",
        "729": "chrome",
        "605": "chrome",
        "604": "chrome",
        "607": "chrome",
        "606": "chrome",
        "601": "chrome",
        "600": "chrome",
        "603": "chrome",
        "602": "chrome",
        "159": "chrome",
        "158": "chrome",
        "112": "chrome",
        "609": "chrome",
        "608": "chrome",
        "976": "opera",
        "634": "chrome",
        "399": "chrome",
        "635": "chrome",
        "959": "safari",
        "958": "safari",
        "398": "chrome",
        "48": "chrome",
        "49": "chrome",
        "951": "safari",
        "950": "safari",
        "953": "safari",
        "952": "safari",
        "42": "chrome",
        "954": "safari",
        "957": "safari",
        "956": "safari",
        "638": "chrome",
        "5": "chrome",
        "639": "chrome",
        "460": "chrome",
        "489": "chrome",
        "488": "chrome",
        "487": "chrome",
        "486": "chrome",
        "485": "chrome",
        "484": "chrome",
        "483": "chrome",
        "482": "chrome",
        "481": "chrome",
        "480": "chrome",
        "509": "chrome",
        "955": "safari",
        "472": "chrome",
        "473": "chrome",
        "470": "chrome",
        "471": "chrome",
        "476": "chrome",
        "477": "chrome",
        "474": "chrome",
        "475": "chrome",
        "478": "chrome",
        "479": "chrome"
    }
}

if __name__ == '__main__':
    ua = 'Chrome'
    # ua = 'random'
    print(choice_brower_model(ua))



import OpenSSL
import OpenSSL._util as pyOpenSSLutil

from scrapy.utils.python import to_unicode


# The OpenSSL symbol is present since 1.1.1 but it's not currently supported in any version of pyOpenSSL.
# Using the binding directly, as this code does, requires cryptography 2.4.
SSL_OP_NO_TLSv1_3 = getattr(pyOpenSSLutil.lib, 'SSL_OP_NO_TLSv1_3', 0)


def ffi_buf_to_string(buf):
    return to_unicode(pyOpenSSLutil.ffi.string(buf))


def x509name_to_string(x509name):
    # from OpenSSL.crypto.X509Name.__repr__
    result_buffer = pyOpenSSLutil.ffi.new("char[]", 512)
    pyOpenSSLutil.lib.X509_NAME_oneline(x509name._name, result_buffer, len(result_buffer))

    return ffi_buf_to_string(result_buffer)


def get_temp_key_info(ssl_object):
    if not hasattr(pyOpenSSLutil.lib, 'SSL_get_server_tmp_key'):  # requires OpenSSL 1.0.2
        return None

    # adapted from OpenSSL apps/s_cb.c::ssl_print_tmp_key()
    temp_key_p = pyOpenSSLutil.ffi.new("EVP_PKEY **")
    if not pyOpenSSLutil.lib.SSL_get_server_tmp_key(ssl_object, temp_key_p):
        return None
    temp_key = temp_key_p[0]
    if temp_key == pyOpenSSLutil.ffi.NULL:
        return None
    temp_key = pyOpenSSLutil.ffi.gc(temp_key, pyOpenSSLutil.lib.EVP_PKEY_free)
    key_info = []
    key_type = pyOpenSSLutil.lib.EVP_PKEY_id(temp_key)
    if key_type == pyOpenSSLutil.lib.EVP_PKEY_RSA:
        key_info.append('RSA')
    elif key_type == pyOpenSSLutil.lib.EVP_PKEY_DH:
        key_info.append('DH')
    elif key_type == pyOpenSSLutil.lib.EVP_PKEY_EC:
        key_info.append('ECDH')
        ec_key = pyOpenSSLutil.lib.EVP_PKEY_get1_EC_KEY(temp_key)
        ec_key = pyOpenSSLutil.ffi.gc(ec_key, pyOpenSSLutil.lib.EC_KEY_free)
        nid = pyOpenSSLutil.lib.EC_GROUP_get_curve_name(pyOpenSSLutil.lib.EC_KEY_get0_group(ec_key))
        cname = pyOpenSSLutil.lib.EC_curve_nid2nist(nid)
        if cname == pyOpenSSLutil.ffi.NULL:
            cname = pyOpenSSLutil.lib.OBJ_nid2sn(nid)
        key_info.append(ffi_buf_to_string(cname))
    else:
        key_info.append(ffi_buf_to_string(pyOpenSSLutil.lib.OBJ_nid2sn(key_type)))
    key_info.append(f'{pyOpenSSLutil.lib.EVP_PKEY_bits(temp_key)} bits')
    return ', '.join(key_info)


def get_openssl_version():
    system_openssl = OpenSSL.SSL.SSLeay_version(
        OpenSSL.SSL.SSLEAY_VERSION
    ).decode('ascii', errors='replace')
    return f'{OpenSSL.version.__version__} ({system_openssl})'



# -*- coding: utf-8 -*-
# 用于过滤重复的item
# @Author  : white.tie
# @Time    : 2022/4/29 14:21
# @File: bloomfilter.py.py
"""
date:2021/10/26
auth:t.y.jie
changed: by zhizhong 2022/5/10

params:: 
from scrapy.cfg :: REDIS_HOST / REDIS_PORT / REDIS_DB / REDIS_PASSWORD <str>
from custom_settings :: Data_Size <int/str>  eg: 1000*10000/百万级/千万级/千万 
from custom_settings :: Aim_Set <dict>  eg: {"title","all_json"} 除重依据字段
"""
import redis
import os
from hashlib import md5
from scrapy.utils.conf import get_config
from bisect import bisect_left
from scrapy.exceptions import DropItem


class SimpleHash(object):
    # cap 代表向量长度
    def __init__(self, cap, seed):
        self.cap = cap
        self.seed = seed

    def hash(self, value):
        ret = 0
        for i in range(len(value)):
            ret += self.seed * ret + ord(value[i])
        # 位运算保证最后的值在0到self.cap之间
        return (self.cap-1) & ret



class BloomFilterPipeline(object):
    def __init__(self, REDIS_HOST, REDIS_PORT, REDIS_DB, REDIS_PASSWORD,REDIS_BF_KEY,REDIS_BIT,Aim_Set):
        """
        :param host: the host of Redis
        :param port: the port of Redis
        :param db: witch db in Redis
        :param blockNum: one blockNum for about 90,000,000; if you have more strings for filtering, increase it.
        :param key: the key's name in Redis
        """
        self.server = redis.Redis(host=REDIS_HOST, port=REDIS_PORT, db=REDIS_DB, password=REDIS_PASSWORD)
        self.bit_size = 1 << REDIS_BIT  # Redis的String类型最大容量为512M，现使用256M
        self.seeds = [5, 7, 11, 13, 31, 37, 61]
        self.key = REDIS_BF_KEY
        self.blockNum = 1
        self.hashfunc = []
        self.aim_set = Aim_Set
        for seed in self.seeds:
            self.hashfunc.append(SimpleHash(self.bit_size, seed))

    @classmethod
    def from_crawler(cls, crawler):
        # 判断运行环境<根据环境变量中是否配置IF_PROD=True,或 测试正式环境settings["IF_PROD"] = True>
        # section_select ="redis_cfg_prod" if os.environ.get('IF_PROD') == "True" or crawler.settings.get('IF_PROD')\
        #      == True or get_config().get("settings","IF_PROD",fallback="False")=="True" else "redis_cfg_dev"
        section_select ="redis_cfg_prod" if os.environ.get('IF_PROD') == "True" \
             or get_config().get("settings","IF_PROD",fallback="False")=="True" else "redis_cfg_dev"
        Data_Size = crawler.settings.get('Data_Size')
        if isinstance(Data_Size,int):
            REDIS_BIT = [25,28,31][bisect_left([101*10000,1000*10000,10000*10000], Data_Size)]
        elif isinstance(Data_Size,str):
            means ='十万' if '十万' in Data_Size else ('百万' if '百万' in Data_Size else ("千万" if "千万" in Data_Size else  "亿"))
            REDIS_BIT = {"十万":24,"百万":25,"千万":28,"亿":31,}.get(means)
        else:
            REDIS_BIT = 28
        return cls(REDIS_HOST=get_config().get(section=section_select,option='REDIS_HOST',fallback='') or crawler.settings.get('REDIS_HOST'),
                REDIS_PORT = get_config().getint(section=section_select,option='REDIS_PORT',fallback='') or crawler.settings.get('REDIS_PORT'),
                REDIS_PASSWORD = get_config().get(section=section_select,option='REDIS_PASSWORD',fallback='')or \
                    crawler.settings.get('REDIS_PASSWORD') or crawler.settings.get('REDIS_PARAMS',{}).get("password"),
                REDIS_DB = get_config().get(section=section_select,option='REDIS_DB',fallback='') or \
                    crawler.settings.get('REDIS_DB') or crawler.settings.get('REDIS_PARAMS',{}).get("db"),
                REDIS_BF_KEY = crawler.spidercls.name,
                REDIS_BIT=REDIS_BIT,
                Aim_Set = crawler.settings.get('Aim_Set',{"title","sources","brand_name", "brand_id", "package", "list_json", "category_id",
                                                 "p_category_id", "contain_json", "all_json",
                                                 "raw_img_url", "raw_pdf_url"})
                )

    def isContains(self, str_input):    
        if not str_input:
            return False
        m5 = md5()
        m5.update(str_input)
        str_input = m5.hexdigest()
        ret = True
        # 获取键名 bloomfilter0
        name = self.key
        for f in self.hashfunc:
            offset = f.hash(str_input)
            # 命令用于对 key 所储存的字符串值，获取指定偏移量上的位(bit)。
            ret = ret & self.server.getbit(name, offset)
        return ret

    def insert(self, str_input):
        m5 = md5()
        m5.update(str_input)
        str_input = m5.hexdigest()
        # name redis键的名称
        name = self.key
        for f in self.hashfunc:
            offset = f.hash(str_input)
            self.server.setbit(name, offset, 1)
            
    def iter_dict_total(self,item_dic):
        """
        @param item_dic: 输入一个字典按照字符串排序
        @return: 输出字典转化为字符串
        """
        item_ls = []
        
        def iter_dict(item_dic):
            # aim_set 目标字段与 item_dic 字段交集
            for key in sorted(self.aim_set&set(item_dic)):
                value = item_dic[key]
                item_ls.append(key)
                if isinstance(value, dict):
                    iter_dict(value)
                elif isinstance(value, list):
                    for value_ in value:
                        iter_dict(value_)
                elif isinstance(value, set):
                    for value_ in value:
                        iter_dict(value_)
                else:
                    item_ls.append(str(value))
                value = None
            return "".join(item_ls).encode()

        return iter_dict(item_dic)
    
    def process_item(self, item, spider):
        postItem = dict(item)
        item_str = self.iter_dict_total(postItem)
        msg = item.get("title", "") or item.get("category_name","") or item.get("url","") or item.get("sources","") # or item.get("")
        if self.isContains(item_str):  # 判断字符串是否存在
            DropItem(f'This is item exists! info: {msg}')
        else:
            self.insert(item_str)
            return item

    def close_spider(self,spider):
        self.server.close()



from unittest import TestCase

from testfixtures import LogCapture

from scrapy.spidermiddlewares.urllength import UrlLengthMiddleware
from scrapy.http import Response, Request
from scrapy.spiders import Spider
from scrapy.utils.test import get_crawler
from scrapy.settings import Settings


class TestUrlLengthMiddleware(TestCase):

    def setUp(self):
        self.maxlength = 25
        settings = Settings({'URLLENGTH_LIMIT': self.maxlength})

        crawler = get_crawler(Spider)
        self.spider = crawler._create_spider('foo')
        self.stats = crawler.stats
        self.mw = UrlLengthMiddleware.from_settings(settings)

        self.response = Response('http://scrapytest.org')
        self.short_url_req = Request('http://scrapytest.org/')
        self.long_url_req = Request('http://scrapytest.org/this_is_a_long_url')
        self.reqs = [self.short_url_req, self.long_url_req]

    def process_spider_output(self):
        return list(self.mw.process_spider_output(self.response, self.reqs, self.spider))

    def test_middleware_works(self):
        self.assertEqual(self.process_spider_output(), [self.short_url_req])

    def test_logging(self):
        with LogCapture() as log:
            self.process_spider_output()

        ric = self.stats.get_value('urllength/request_ignored_count', spider=self.spider)
        self.assertEqual(ric, 1)

        self.assertIn(f'Ignoring link (url length > {self.maxlength})', str(log))
